/**
 * Component can be used with movable to chase after a target, attempts to
 * avoid obstacles by leaving a trail behind the target.
 */
bento.define('components/chaser', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var PI = Math.PI;
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var movable;
        var targets = [];
        var rotationTarget = 0;
        var diff;
        var rotationSpeed = PI / 50;
        var spawnTarget = function () {
            var target = new Entity({
                z: 0,
                name: 'subTarget',
                position: component.target.position.clone(),
            }).attach({
                update: function () {
                    // collided with entity
                    if (target.collidesWith(entity, new Vector2(0, 0), function () {
                        Bento.objects.remove(target);
                        targets.shift();
                    }));
                }
            });
            target.dimension = new Rectangle(-8, -8, 16, 16);
            targets.push(target);
            Bento.objects.attach(target);
        };
        var searchClosest = function () {
            var i,
                diff,
                targetDiff,
                prev,
                other,
                otherDiff,
                playerDiff,
                target = targets[0],
                totalDist = 0,
                getClosest = function () {
                    // shifts until the first target is other
                    while (i--) {
                        Bento.objects.remove(targets.shift());
                    }
                    return other;
                };

            // should calculate the total distance the entity has to travel
            // and compare that to distance it can by taking a shortcut
            for (i = 1; i < targets.length; ++i) {
                other = targets[i];
                prev = targets[i - 1];
                diff = other.position.substract(prev.position);
                totalDist += diff.x * diff.x + diff.y * diff.y;
            }
            //
            // diff = entity.position.substract(target.position);
            // targetDiff = diff.x * diff.x + diff.y * diff.y;
            targetDiff = totalDist;

            // try player
            diff = entity.position.substract(component.target.position);
            playerDiff = diff.x * diff.x + diff.y * diff.y;
            if (playerDiff < targetDiff) {
                // reset targets and return player as target
                resetTargets();
                return component.target;
            }

            // try other targets
            for (i = 1; i < targets.length; ++i) {
                other = targets[i];
                diff = entity.position.substract(other.position);
                otherDiff = diff.x * diff.x + diff.y * diff.y;
                // is this one closer?
                if (otherDiff < targetDiff) {
                    return getClosest();
                }
            }
            // nothing is closer
            return target;
        };
        var resetTargets = function () {
            while (targets.length) {
                Bento.objects.remove(targets.pop());
            }
        };
        var component = {
            name: 'chaser',
            timer: 0,
            target: settings.target || null,
            rotation: PI / 2,
            subTarget: null,
            chasing: true,
            speed: settings.speed || 1,
            bump: false,
            start: function (data) {},
            destroy: function (data) {
                resetTargets();
            },
            update: function (data) {
                this.timer += 1;
                if (!this.chasing) {
                    return;
                }
                if (!movable) {
                    entity.getComponent('movable', function (obj) {
                        movable = obj;
                    });
                    if (movable) {
                        movable.collisionListeners.push(function (data) {
                            // component.bump = true;
                            // Utils.timeout(10, function () {
                            //     component.bump = false;
                            // });
                        });
                    }
                    return;
                }
                // chase after targets
                if (!targets.length) {
                    spawnTarget();
                }

                this.subTarget = searchClosest();

                // spawn targets every second
                if (this.timer % 30 === 0) {
                    spawnTarget();
                }

                // look at target
                rotationTarget = Math.atan2(
                    this.subTarget.position.y - entity.position.y,
                    this.subTarget.position.x - entity.position.x
                );

                // change range in 0 - 2PI
                rotationTarget = (rotationTarget + 2 * PI) % (2 * PI);

                // rotate the shortest way
                diff = rotationTarget - this.rotation;
                if (diff > PI) {
                    this.rotation = Utils.approach(this.rotation, rotationTarget - 2 * PI, rotationSpeed);
                } else if (diff < -PI) {
                    this.rotation = Utils.approach(this.rotation, rotationTarget + 2 * PI, rotationSpeed);
                } else {
                    this.rotation = Utils.approach(this.rotation, rotationTarget, rotationSpeed);
                }
                // change range in 0 - 2PI
                this.rotation = (this.rotation + 2 * PI) % (2 * PI);

                //entity.rotation = rotation;

                // entity.position.x += speed * Math.cos(rotation);
                // entity.position.y += speed * Math.sin(rotation);

                if (!this.bump) {
                    movable.speed.x = Utils.approach(movable.speed.x, this.speed * Math.cos(this.rotation), 0.1);
                    movable.speed.y = Utils.approach(movable.speed.y, this.speed * Math.sin(this.rotation), 0.1);
                }

                // flip
                if (movable.speed.x < 0) {
                    entity.scale.x = 1;
                } else {
                    entity.scale.x = -1;
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };

        return component;
    };
});