/**
 * Draggable behavior for cards
 */
bento.define('components/draggable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/clickable'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Clickable
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     * @param {Function} settings.dragInit // user touched the screen
     * @param {Function} settings.dragStart // the dragging has started (if already dragging, this isnt called)
     * @param {Function} settings.dragEnd // user released, fires a dragEnd event
     */
    var dragEntity = null;
    var dragId = -1;
    var distance = new Vector2(0, 0);

    // global event listener: when screens are hidden, always reset draggables
    EventSystem.on('screenHidden', function () {
        dragEntity = null;
        dragId = -1;
    });
    var module = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var startPosition = new Vector2(0, 0);
        var diffPosition = new Vector2(0, 0);
        var dropTargets = Bento.objects.getByFamily('dropTargets');
        var activated = true;
        var clickable = new Clickable({
            onClick: function (evt) {
                var data = {};
                if (settings.dragInit) {
                    settings.dragInit();
                }
                // already dragging
                if (dragEntity || !component.canDrag) {
                    // dragEntity = null;
                    return;
                }
                dragEntity = entity;
                dragId = evt.id;
                diffPosition = entity.position.subtract(evt.localPosition);
                startPosition = entity.position.clone();
                distance = new Vector2(0, 0);
                data = {
                    entity: dragEntity
                };
                if (settings.dragStart) {
                    settings.dragStart(data);
                }
                EventSystem.fire('dragStart', data);
            },
            pointerMove: function (evt) {
                if (dragEntity === entity && evt.id === dragId) {
                    entity.position = evt.localPosition.add(diffPosition);
                    distance = entity.position.subtract(startPosition);
                }
            },
            pointerUp: function (evt) {
                var data;
                var endPosition;
                if (dragEntity !== entity || evt.id !== dragId) {
                    return;
                }
                endPosition = entity.position.clone();
                // snap back
                dragEntity.position = startPosition.clone();

                data = {
                    entity: dragEntity,
                    startPosition: startPosition,
                    endPosition: endPosition,
                    event: evt
                };
                // send event to alert droptargets
                EventSystem.fire('dragEnd', data);

                if (settings.dragEnd) {
                    settings.dragEnd(data);
                }

                // release
                dragEntity = null;
                dragId = -1;
                distance = new Vector2(0, 0);
            }
        });
        var component = {
            name: 'draggable',
            canDrag: true,
            attached: function (data) {
                entity = data.entity;
                // add the clickable component
                entity.attach(clickable);
                this.pointers = Bento.input.getPointers();
            },
            getStartPosition: function () {
                return startPosition;
            },
            activate: function () {
                if (activated) {
                    return;
                }
                activated = true;
                entity.attach(clickable);
            },
            deactivate: function () {
                if (!activated) {
                    return;
                }
                activated = false;
                entity.remove(clickable);
            },
            pointers: [],
            update: function () {
                if (dragEntity !== null && this.pointers.length === 0) {
                    // missed a pointerUp somehow
                    dragEntity = null;
                }
            }
        };

        return component;
    };

    // static
    module.isDragging = function () {
        return dragEntity !== null;
    };
    module.getDragEntity = function () {
        return dragEntity;
    };
    module.getDistance = function () {
        return distance;
    };

    return module;
});