/**
 * The movable component moves the parent entity constantly according its speed
 * It will also stop movement if it hits an object with family 'solid'
 */
bento.define('components/movable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var solids = Bento.objects.getByFamily('solid'),
        shouldIgnore = function (ignoredArray, familyArray) {
            var i, family;

            if (!familyArray) {
                return false;
            }

            for (i = 0; i < familyArray.length; ++i) {
                family = familyArray[i];
                if (ignoredArray.indexOf(family) >= 0) {
                    return true;
                }
            }
            return false;
        },
        component = function (settings) {
            this.name = 'movable';
            this.paused = false;
            this.speed = settings.speed || new Vector2(0, 0);
            this.maxSpeed = settings.maxSpeed || new Vector2(Infinity, Infinity);
            this.collided = [];
            this.collidedEntity = null;
            this.entity = null;
            this.tileColliders = null;
            this.ignorePolygons = true;
            this.settings = settings || {};
            this.gravity = this.settings.gravity || 0;
            this.settings.friction = this.settings.friction || new Vector2(1, 1);
            this.settings.bounce = this.settings.bounce || new Vector2(0, 0);
            this.settings.ignoreSolids = this.settings.ignoreSolids || false;
            this.settings.onCollide = this.settings.onCollide || function () {};
            this.settings.useHshg = this.settings.useHshg || false;
            this.settings.ignoredFamilies = this.settings.ignoredFamilies || [];
            this.collisionListeners = [];
        };
    component.prototype.update = function (data) {
        var entity = this.entity,
            speed = this.speed,
            normalizedSpeed = speed.clone().normalize(),
            collided = this.collided,
            otherBox,
            settings = this.settings,
            other,
            self = this,
            onCollide = function () {
                var i,
                    box,
                    diffBox,
                    direction = new Vector2(Utils.sign(speed.x), Utils.sign(speed.y)),
                    pushX = function (dir) {
                        // pushes out entity in x
                        // dir = -1 or 1
                        entity.position.x += diffBox.width * dir;
                    },
                    pushY = function (dir) {
                        // pushes out entity in y
                        // dir = -1 or 1
                        entity.position.y += diffBox.height * dir;
                    },
                    isColliding = function () {
                        return other.collidesWith(entity);
                    },
                    newSpeed = function (direction) {
                        // calculates new speed: bounce in normal direction, friction in perpendicular
                        // direction should be 'x' or 'y'
                        var perpendicular = (direction === 'x' ? 'y' : 'x');
                        speed[direction] *= -settings.bounce[direction];
                        speed[perpendicular] *= settings.friction[perpendicular];

                    };

                // direction should not be 0
                direction.x = direction.x || 1;
                direction.y = direction.y || 1;

                // collect difference rectangle 
                for (i = 0; i < collided.length; ++i) {
                    other = collided[i];
                    otherBox = other.getBoundingBox();

                    // polygon?
                    if (otherBox.isPolygon) {
                        // push out in reverse direction until not colliding
                        entity.position.x -= normalizedSpeed.x;
                        entity.position.y -= normalizedSpeed.y;
                        while (isColliding()) {
                            entity.position.x -= normalizedSpeed.x;
                            entity.position.y -= normalizedSpeed.y;
                        }
                        // temp: reverse direction
                        // todo: find out which direction the bounce really is
                        speed['x'] *= -settings.bounce['x'];
                        speed['y'] *= -settings.bounce['y'];
                    } else {
                        // normal rectangle collision
                        diffBox = entity.getBoundingBox().intersection(otherBox);
                        if (diffBox.width === 0 && diffBox.height === 0) {
                            // not colliding?
                            // hshg is wrong apparently?
                            continue;
                        }

                        // push out of x or y?
                        // push out based on box size
                        if (diffBox.width < diffBox.height) {
                            pushX(-direction.x);
                            // check
                            if (isColliding()) {
                                // failed!
                                // revert x and try y instead
                                pushX(direction.x);
                                pushY(-direction.y);
                                //  
                                newSpeed('y');
                            } else {
                                newSpeed('x');
                            }
                        } else {
                            pushY(-direction.y);
                            if (isColliding()) {
                                pushY(direction.y);
                                pushX(-direction.x);
                                newSpeed('x');
                            } else {
                                newSpeed('y');
                            }
                        }
                    }
                }
                // collision with solid 
                settings.onCollide({
                    entity: self.entity,
                    others: collided,
                    other: other
                });

                for (i = 0; i < self.collisionListeners.length; ++i) {
                    self.collisionListeners[i]({
                        entity: self.entity,
                        others: collided,
                        other: other
                    });
                }
            };

        if (this.paused) {
            return;
        }

        // gravity
        if (this.gravity) {
            speed.y += this.gravity * data.speed;
        }

        // cap speed
        if (this.maxSpeed) {
            speed.x = Utils.clamp(-this.maxSpeed.x, speed.x, this.maxSpeed.x);
            speed.y = Utils.clamp(-this.maxSpeed.y, speed.y, this.maxSpeed.y);
        }

        // move entity
        entity.position.x += speed.x * data.speed;
        entity.position.y += speed.y * data.speed;

        // check if collided
        if (this.tileColliders && !settings.ignoreSolids) {
            this.tileColliders.collidesWith(entity, new Vector2(0, 0), function (other) {
                self.collidedEntity = other;
                collided.push(other);
                if (settings.onCollideHash) {
                    settings.onCollideHash(other);
                }
            });
        } else {
            Bento.objects.get('tileColliders', function (obj) {
                self.tileColliders = obj;
            });
        }
        // collision with solids
        if (!this.settings.useHshg && !settings.ignoreSolids) {
            entity.collidesWithGroup(solids, new Vector2(0, 0), function (other) {
                // fine collision
                var otherBox = other.getBoundingBox();
                // is this an ignored object?
                if (shouldIgnore(self.settings.ignoredFamilies, other.family)) {
                    return;
                }

                if (otherBox.isPolygon) {
                    // default behavior: ignore polygons because they cause bugs
                    if (self.ignorePolygons) {
                        return;
                    }
                    if (!otherBox.intersect(entity.getBoundingBox())) {
                        return;
                    }
                }

                self.collidedEntity = other;
                self.collided.push(other);
            });
        }

        if (collided.length) {
            onCollide();
        } else {}

        // reset collided
        this.collided.length = 0;
        this.collidedEntity = null;
    };
    component.prototype.onParentCollided = function (data) {
        var otherBox,
            other = data.other;
        if (this.settings.ignoreSolids || !this.settings.useHshg) {
            return;
        }
        // is this an ignored object?
        if (shouldIgnore(this.settings.ignoredFamilies, other.family)) {
            return;
        }

        // is of family solid?
        if (data.other.family.indexOf('solid') !== -1) {
            // fine collision
            otherBox = data.other.getBoundingBox();
            if (otherBox.isPolygon) {
                // default behavior: ignore polygons because they cause bugs
                if (this.ignorePolygons) {
                    return;
                }
                if (!otherBox.intersect(data.entity.getBoundingBox())) {
                    return;
                }
            }

            this.collidedEntity = data.other;
            this.collided.push(data.other);
        }
    }
    component.prototype.attached = function (data) {
        var self = this;
        this.entity = data.entity;
        if (!this.tileColliders) {
            Bento.objects.get('tileColliders', function (obj) {
                self.tileColliders = obj;
            });
        }

    }
    return component;
});