/**
 * Walk component for critters
 */
bento.define('components/walk', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'entities/particles/death',
    'entities/effect'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Tween,
    DeathParticle,
    Effect
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity; // the corresponding critter
        var isWalking = false;
        var groundY = 0;
        var grounded = true;
        var gravity = 0.5;
        var speedY = 0;
        var speedX = 2;
        var collectScale = 1;
        var bounceScale = new Vector2(1, 1);
        var idleScale = new Vector2(1, 1);
        var twitterScale = new Vector2(1, 1);
        var bounceStrength = 0.2;
        var directionScale = new Vector2(1, 1);
        var state = 'idle';
        var wiggleTween = null;
        /**
         *  Idle state, can walk around
         */
        var idle = function (data) {
            if (entity.position.x !== component.targetX) {
                // turn
                if (entity.position.x < component.targetX) {
                    directionScale.x = 1;
                } else {
                    directionScale.x = -1;
                }

                // move
                entity.position.x = Utils.approach(entity.position.x, component.targetX, speedX * data.speed);
                isWalking = true;

                // little jump
                if (grounded) {
                    speedY = -3;
                }
            } else {
                isWalking = false;
                directionScale.x = component.defaultScaleX;
            }

            // jumpy walk!
            if (isWalking || !grounded) {
                speedY += gravity * data.speed;
                entity.position.y += speedY * data.speed;

            }

            if (entity.position.y < groundY) {
                grounded = false;
            }

            if (entity.position.y >= groundY) {
                entity.position.y = groundY;
                // land
                if (!grounded) {
                    grounded = true;

                    new Tween({
                        from: 0,
                        to: 1,
                        'in': 60,
                        ease: 'elastic',
                        alpha: -5,
                        beta: 3,
                        'do': function (v, t) {
                            bounceScale.x = 1 + bounceStrength - v * bounceStrength;
                            bounceScale.y = 1 - bounceStrength + v * bounceStrength;
                        }
                    });
                }
            }

            // idle
            idleScale.x = 1 + 0.05 * Math.cos(entity.timer / 15);
            idleScale.y = 1 - 0.05 * Math.cos(entity.timer / 15);
        };
        /**
         *  JumpAttack state
         */
        var jumpAttack = function (settings) {
            var targetX = settings.targetX;
            var target = settings.other;
            var targetXOffset = targetX - settings.other.position.x;
            var startX = entity.position.x;
            var other = settings.other;
            var onHit = settings.onHit;
            var onFinished = settings.onFinished;
            var jumpTime = 25;
            if (state !== 'idle') {
                console.log('WARNING: should be in idle state');
                // return;
            }
            if (entity.card.getProperty('currentHp') <= 0) {
                // critter is dead...
                return;
            }
            state = 'attack';

            Bento.audio.stopSound('sfx_jump');
            Bento.audio.playSound('sfx_jump');

            // put critter on the ground
            entity.position.y = groundY;
            idleScale.x = 1;
            idleScale.y = 1;
            directionScale.x = component.defaultScaleX;

            // jump
            new Tween({
                from: entity.position.y,
                to: groundY - 32,
                'in': jumpTime / 2,
                ease: 'cuberoot',
                'do': function (v, t) {
                    entity.position.y = v;
                },
                onComplete: function () {
                    new Tween({
                        from: entity.position.y,
                        to: groundY,
                        'in': jumpTime / 2,
                        ease: 'cubic',
                        'do': function (v, t) {
                            entity.position.y = v;
                        },
                        onComplete: function () {
                            entity.position.y = groundY;
                        }
                    });
                }
            });

            // attack rotation
            new Tween({
                from: 0,
                to: -Math.PI / 6 * component.defaultScaleX,
                'in': jumpTime / 2,
                ease: 'exponential',
                alpha: -3,
                'do': function (v, t) {
                    entity.rotation = v;
                },
                onComplete: function () {
                    // console.log('slowmo')
                    // Bento.setGameSpeed(0.25);
                    Utils.TapticEngine.impact({
                        style: 'light'
                    });

                    new Tween({
                        from: entity.rotation,
                        to: Math.PI / 6 * component.defaultScaleX,
                        'in': jumpTime - jumpTime / 2,
                        ease: 'exponential',
                        alpha: -1,
                        'do': function (v, t) {
                            entity.rotation = v;
                        },
                        onComplete: function () {
                            // TODO: critter may have died in the meantime
                            
                            if (onHit) {
                                onHit(entity, other);
                            }
                            state = 'wait';
                            new Tween({
                                from: entity.rotation,
                                to: 0,
                                'in': 30,
                                ease: 'elastic',
                                alpha: -4,
                                beta: 1,
                                'do': function (v, t) {
                                    entity.rotation = v;
                                },
                                onComplete: function () {
                                    if (state !== 'wait') {
                                        // if (state === 'death') {
                                        // critter may have died in the meantime
                                        return;
                                    }
                                    state = 'idle';
                                    // end of attack
                                    if (onFinished) {
                                        onFinished(entity, other);
                                    }
                                    // component.targetX = endPositionX;
                                    // speedX = 3;
                                }
                            });
                            entity.rotation = 0;
                        }
                    });
                }
            });

            // move to target
            new Tween({
                from: 0,
                to: 1,
                'in': jumpTime,
                ease: 'linear',
                'do': function (v, t) {
                    entity.position.x = startX + v * (target.position.x + targetXOffset - startX);
                },
                onComplete: function () {

                }
            });
        };
        var component = {
            name: 'walk',
            targetX: 0,
            defaultScaleX: settings.defaultScaleX || 1,
            speed: new Vector2(0, 0),
            start: function (data) {
                // twitter?
                if (settings.isTwitter) {
                    this.defaultScaleX = -1;
                }
            },
            update: function (data) {
                entity.scale.x = bounceScale.x * directionScale.x * idleScale.x * collectScale * twitterScale.x;
                entity.scale.y = bounceScale.y * idleScale.y * collectScale * twitterScale.y;

                if (state === 'idle' || state === 'run') {
                    idle(data);
                }

                // if (grounded && !isWalking) {} else {
                //     idleScale.x = 1;
                //     idleScale.y = 1;
                // }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                groundY = entity.position.y;
            },
            jumpAttack: jumpAttack,
            isIdling: function () {
                return state === 'idle';
            },
            isAttacking: function () {
                return state === 'attack';
            },
            death: function (targetX) {
                var duration = 36;
                var shownEffect = false;
                var particle;
                state = 'death';
                // show death animation (if in screen)
                if (entity.position.x > 0 && entity.position.x < viewport.width) {
                    // fly off
                    new Tween({
                        from: entity.position.x,
                        to: targetX,
                        'in': duration,
                        ease: 'linear',
                        'do': function (v, t) {
                            entity.position.x = v;
                            if ((entity.position.x < 0 || entity.position.x > viewport.width) && !shownEffect) {
                                shownEffect = true;
                                particle = new DeathParticle({
                                    position: entity.position.clone()
                                });
                                particle.position.x = Utils.clamp(0, entity.position.x, viewport.width);
                                if (particle.position.x <= viewport.x + viewport.width / 2) {
                                    particle.rotation = Math.PI / 2;
                                } else {
                                    particle.rotation = -Math.PI / 2;
                                }
                                entity.visible = false;
                                Bento.audio.stopSound('sfx_death');
                                Bento.audio.playSound('sfx_death');
                            }
                        }
                    });
                    new Tween({
                        from: entity.position.y,
                        to: entity.position.y - 80 - Utils.getRandom(40),
                        'in': duration,
                        ease: 'squareroot',
                        'do': function (v, t) {
                            entity.position.y = v;
                        }
                    });
                    new Tween({
                        from: entity.rotation,
                        to: entity.rotation - Math.PI / 2,
                        'in': duration,
                        ease: 'linear',
                        'do': function (v, t) {
                            entity.rotation = v;
                        }
                    });
                } else {
                    // jsut make invisible?
                    entity.visible = false;
                }

            },
            jump: function (jumpStrength) {
                jumpStrength = jumpStrength || 4;
                speedY = -jumpStrength;
                grounded = false;
            },
            getHit: function () {
                new Tween({
                    from: Math.PI / 8 * component.defaultScaleX,
                    to: 0,
                    'in': 60,
                    ease: 'elastic',
                    alpha: -5,
                    beta: 3,
                    'do': function (v, t) {
                        entity.rotation = v;
                    }
                });
            },
            getUpAndRun: function (settings) {
                var self = this;
                state = 'run';
                speedY = -5;
                // setup
                entity.visible = true;
                entity.setOriginRelative(new Vector2(0.5, 1));
                entity.position.x = viewport.width / 2 - entity.dimension.width / 2;
                entity.rotation = Math.PI / 2;

                if (wiggleTween) {
                    wiggleTween.stop();
                }

                Bento.audio.playSound('sfx_jump');

                new Tween({
                    from: entity.rotation,
                    to: 0,
                    'in': 30,
                    ease: 'easeOutBack',
                    'do': function (v, t) {
                        entity.rotation = v;
                    },
                    onComplete: function () {
                        Utils.timeout(30, function () {
                            self.targetX = viewport.width * 2;
                            speedX *= 2;
                            gravity *= 2;
                            new Tween({
                                from: 0,
                                to: Math.PI / 10,
                                'in': 60,
                                ease: 'sin',
                                alpha: 6,
                                'do': function (v, t) {
                                    if (t % 4 === 0) {
                                        new Effect({
                                            sprite: new Sprite(Bento.assets.getJson('animations/poof')),
                                            position: entity.position.add(new Vector2(-4, 0)),
                                            originRelative: new Vector2(0.5, 0.75)
                                        });
                                    }
                                    if (t % 10 === 0) {
                                        Bento.audio.stopSound('sfx_battle_run');
                                        Bento.audio.playSound('sfx_battle_run');
                                    }
                                    entity.rotation = v;
                                },
                                onComplete: function () {
                                    if (settings.hasEscaped) {
                                        settings.hasEscaped();
                                    }
                                }
                            });
                        });
                    }
                });
            },
            deadWiggle: function () {
                var time = 60;
                var side = Math.PI / 2;
                var wiggle = function () {
                    var moveTo = function () {
                        wiggleTween = new Tween({
                            from: entity.rotation,
                            to: side - Math.PI / (6 + Utils.getRandom(3)),
                            'in': 5 + Utils.getRandom(5),
                            ease: 'linear',
                            'do': function (v, t) {
                                entity.rotation = v;
                            },
                            onComplete: function () {
                                if (state !== 'wiggle') {
                                    return;
                                }
                                Utils.timeout(Utils.getRandom(2) * 10, drop);
                            }
                        });
                    };
                    var drop = function () {
                        wiggleTween = new Tween({
                            from: entity.rotation,
                            to: side,
                            'in': 60,
                            ease: 'elastic',
                            alpha: -3 - Utils.getRandom(5),
                            beta: -2 - Utils.getRandom(4),
                            'do': function (v, t) {
                                entity.rotation = v;
                            },
                            onComplete: function () {
                                if (state !== 'wiggle') {
                                    return;
                                }
                                randomizeTime();
                                Utils.timeout(time, wiggle);
                            }
                        });
                    };
                    if (state !== 'wiggle') {
                        return;
                    }
                    moveTo();
                };
                var randomizeTime = function () {
                    time = 30 + Utils.getRandom(120);
                };
                // setup
                state = 'wiggle';
                entity.visible = true;
                if (entity.dimension.width > entity.dimension.height) {
                    // wide enemies should go upside down
                    entity.setOriginRelative(new Vector2(0.5, 0.2));
                    side = Math.PI;
                } else {
                    entity.setOriginRelative(new Vector2(0.2, 0.5));
                }
                entity.position.x = viewport.width / 2;
                entity.position.y = groundY;
                entity.rotation = side;
                directionScale.x = -1;

                randomizeTime();
                Utils.timeout(time, wiggle);
            },
            beamUp: function (settings) {
                var beamUpHalf = function () {
                    new Tween({
                        from: entity.position.y,
                        to: entity.position.y + (settings.target - entity.position.y) * 2 / 3,
                        'in': 140,
                        ease: 'linear',
                        'do': function (v, t) {
                            entity.position.y = v;
                        },
                        onComplete: function () {
                            // success or fail?
                            if (settings.success) {
                                beamUpFull();
                            } else {
                                drop();
                            }
                        }
                    });
                };
                var drop = function () {
                    new Tween({
                        from: entity.position.y,
                        to: groundY,
                        'in': 60,
                        ease: 'easeOutBounce',
                        'do': function (v, t) {
                            entity.position.y = v;
                            if (t === 20) {
                                Bento.audio.playSound('sfx_drop');
                            }
                        },
                        onComplete: function () {
                            settings.onComplete();
                            //     new Tween({
                            //         from: entity.rotation + Math.PI / 8,
                            //         to: entity.rotation,
                            //         'in': 20,
                            //         ease: 'elastic',
                            //         alpha: -5,
                            //         beta: -3,
                            //         'do': function (v, t) {
                            //             entity.rotation = v;
                            //         },
                            //         onComplete: settings.onComplete
                            //     });
                        }
                    });
                };
                var beamUpFull = function () {
                    new Tween({
                        from: entity.position.y,
                        to: settings.target,
                        'in': 50,
                        ease: 'linear',
                        'do': function (v, t) {
                            entity.position.y = v;
                        },
                        onComplete: settings.onComplete
                    });
                    new Tween({
                        from: 1,
                        to: 0,
                        'in': 50,
                        ease: 'linear',
                        'do': function (v, t) {
                            collectScale = v;
                        }
                    });
                };
                beamUpHalf();
            }
        };
        return component;
    };
});