/**
 * Battle manager
 */
bento.define('entities/battle/manager', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/canvas',
    'bento/gui/counter',
    'bento/gui/clickbutton',
    'entities/battle/critter',
    'entities/card',
    'entities/effect',
    'components/walk',
    'components/enemy',
    'gui/damagecounter',
    'gui/hpbar',
    'bento/gui/text',
    'entities/particles/hit',
    'entities/battle/catchminigame',
    'entities/battle/catchpopup',
    'entities/battle/caughtpopup',
    'entities/battle/gameover',
    'gui/storydialog',
    'gui/fade',
    'entities/battle/attackbutton',
    'entities/battle/coin',
    'modules/zoom',
    'entities/battle/minigames/base',
    'modules/playtimer'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Canvas,
    Counter,
    ClickButton,
    Critter,
    Card,
    Effect,
    Walk,
    Enemy,
    DamageCounter,
    HpBar,
    Text,
    HitParticle,
    CatchMinigame,
    CatchPopup,
    CaughtPopup,
    GameOver,
    StoryDialog,
    Fade,
    AttackButton,
    Coin,
    Zoom,
    MiniGame,
    PlayTimer
) {
    'use strict';
    return function (settings) {
        /*settings = {
            grid: Grid // reference to grid
            data: Object // passed from battle screen,
            isBossBattle: Boolean
        }*/
        var viewport = Bento.getViewport();
        // settings
        var playerCritters = Utils.getTeam();
        var enemyCritters = settings.enemies || [];
        var boss = settings.data.boss;
        var isBossBattle = settings.isBossBattle;
        var isTwitterBattle = settings.isTwitterBattle;
        // active entities
        var critters = [];
        var enemies = [];
        var frontCritter = null;
        var frontEnemy = null;
        var canAttack = true;
        var isAnyAttacking = false;
        var isEnemyAttacking = false;
        var currentSpecialAttack = null;
        var currentTurn = 0;
        var completedSpecialAttacksInARow = 0;
        // time at which battle started, according to playTimer
        var battleStartTime = 0;
        // time at which last battle started
        var timeOfPrevBattleStart = Bento.saveState.load('timeOfPrevBattleStart', PlayTimer.playTimer);
        var newCritter = false;
        var manager = new Entity({
            z: 0,
            name: 'battleManager'
        }).extend({
            isBossBattle: settings.isBossBattle,
            isTwitterBattle: settings.isTwitterBattle,
            state: 'battle', // 'victory', 'defeat'
            /**
             * Get front critter
             */
            getFrontCritter: function () {
                return frontCritter;
            },
            /**
             * Get front enemy
             */
            getFrontEnemy: function () {
                return frontEnemy;
            },
            critterDeath: function (critter, isEnemy) {
                var i;
                // disable card
                critter.card.getAvatar().alpha = 0.5;
                if (critter.card.abilityButton) {
                    critter.card.abilityButton.setActive(false);
                }
                // unable to walk
                critter.getComponent('walk', function (walkComponent) {
                    walkComponent.death(isEnemy ? critter.position.x + viewport.width : critter.position.x - viewport.width);
                });

                if (critter === frontCritter) {
                    // currently the front critter: find the next critter and put that forward
                    for (i = 0; i < critters.length; ++i) {
                        if (critters[i].getProperty('currentHp') > 0) {
                            frontCritter = critters[i];
                            frontCritter.getComponent('walk', function (walkComponent) {
                                walkComponent.targetX = 32;
                            });
                            return;
                        }
                    }
                    // no more critters?
                    gameOver();
                    sendBattleTime();
                } else if (critter === frontEnemy) {
                    // currently the front enemies: find the next critter and put that forward
                    for (i = 0; i < enemies.length; ++i) {
                        if (enemies[i].getProperty('currentHp') > 0) {
                            frontEnemy = enemies[i];
                            frontEnemy.getComponent('walk', function (walkComponent) {
                                walkComponent.targetX = viewport.width - 32;
                            });
                            frontEnemy.visible = true;
                            return;
                        }
                    }
                    // no more enemies?
                    manager.state = 'victory';
                    // Bento.audio.setVolume(0.5, 'bgm_battle');
                    sendBattleTime();
                    // show victory
                    Utils.timeout(60, function () {
                        if (!isTwitterBattle) {
                            Utils.saveTeam();
                        }
                        new GameOver({
                            name: 'victory',
                            onClosed: function () {
                                // count enemies defeated
                                var planetName = Utils.worldMap.planetData.planetName;
                                var defeated = Bento.saveState.load(planetName + '-enemyTotal', 0);
                                Bento.saveState.save(planetName + '-enemyTotal', defeated + 1);

                                manager.catchEnemy();
                                frontCritter.getComponent('walk', function (walkComponent) {
                                    walkComponent.targetX = -viewport.width / 2;
                                });
                                // dead wiggle
                                frontEnemy.getComponent('walk', function (walkComponent) {
                                    walkComponent.targetX = viewport.width / 2;
                                    walkComponent.deadWiggle();
                                });
                            }
                        });
                        Bento.objects.get('battlePauseButton', function (battlePauseButton) {
                            Bento.objects.remove(battlePauseButton);
                        });

                    });
                }
            },
            getGroundY: function () {
                return groundY;
            },
            /**
             * Call this when one critter does damage to another
             * @param {Entity} self - the critter doing the damage
             * @param {Entity} other - the critter receiving the damage
             * @param {Number} damageBoost - Multiplier for damage, 1 by default
             * @param {Function} onDeath - Callback if the critter was killed during the attack
             */
            onHit: function (self, other, damageBoost, onDeath) {
                var enemyBehavior = other.getComponent('enemyBehavior');
                var position = other.card.getWorldPosition();
                var attack = 0;
                if (manager.state !== 'battle') {
                    return;
                }
                damageBoost = damageBoost || 1;

                // compare elements
                damageBoost *= compareElements(self.getProperty('element'), other.getProperty('element'));

                attack = Math.ceil(self.getProperty('attack') * damageBoost);

                other.getComponent('walk', function (component) {
                    component.getHit();
                });
                new HitParticle({
                    position: new Vector2(
                        other.position.x + 32,
                        other.position.y - other.dimension.height / 2
                    )
                }).scale.x = -1;
                // damageCounter on critter and corresponding card
                new DamageCounter({
                    value: -attack,
                    position: new Vector2(
                        other.position.x,
                        other.position.y - other.dimension.height
                    )
                });
                // show at corresponding carc
                new DamageCounter({
                    value: -attack,
                    position: position.add(new Vector2(0, -8))
                });
                new Effect({
                    name: 'normalHit',
                    sprite: new Sprite(Bento.assets.getJson('animations/slash')),
                    position: position.add(new Vector2(0, 0)),
                    originRelative: new Vector2(0.5, 0.5)
                });


                // enemy sound
                if (self === frontEnemy) {
                    Bento.audio.stopSound('sfx_battle_bump');
                    Bento.audio.playSound('sfx_battle_bump');
                }

                // deplete currentHp
                other.addToProperty('currentHp', -attack);
                other.card.addHp(-attack);

                // critter died?
                if (other.getProperty('currentHp') <= 0) {
                    if (enemyBehavior) {
                        manager.critterDeath(other, true);
                        // TODO: drop amount of coins depending on critter property
                        dropCoins(other, other.getProperty('coins') || 0);
                    } else {
                        manager.critterDeath(other, false);
                    }
                    if (onDeath) {
                        onDeath();
                    }
                }

                // save game immediately (to prevent cheating)
                if (!isTwitterBattle) {
                    Utils.saveTeam();
                }
            },
            /**
             * Ask player to catch the enemy
             */
            catchEnemy: function () {
                var i;
                var hasRetried = false;
                var doRetry = function () {
                    hasRetried = true;
                    new Tween({
                        from: popup.position.x,
                        to: viewport.width / 2 - 2,
                        'in': 30,
                        ease: 'easeOutExpo',
                        'do': function (v, t) {
                            popup.position.x = v;
                        },
                        onComplete: function () {
                            var price = Bento.assets.getJson('creatures/retrycost')[frontEnemy.getProperty('rarity') || 10000];
                            extraChance = 20; // ???
                            popup.addChance(extraChance, price);
                        }
                    });
                };
                var popup = new CatchPopup({
                    chance: frontEnemy.getProperty('catchChance'),
                    position: new Vector2(viewport.width * 2, viewport.height / 5),
                    onExit: runAway,
                    onCollect: function () {
                        // remove popup
                        popup.exit(function () {
                            if (hasRetried) {
                                catchSequence(runAway);
                            } else {
                                catchSequence(doRetry, extraChance);
                            }
                            // new CatchMinigame({
                            //     manager: manager,
                            //     frontEnemy: frontEnemy,
                            //     difficulty: frontEnemy.getProperty('catchChance'),
                            //     onComplete: function (success) {
                            //         if (success) {
                            //             catchEnemy();
                            //             victory({
                            //                 caught: frontEnemy
                            //             });
                            //         } else {
                            //             runAway();
                            //         }
                            //     }
                            // });
                        });
                    }
                });
                // hide enemy cards
                for (i = 0; i < enemies.length; ++i) {
                    enemies[i].card.visible = false;
                }
                // popup question
                Bento.objects.attach(popup);
                new Tween({
                    from: popup.position.x,
                    to: viewport.width / 2 - 2,
                    'in': 30,
                    ease: 'easeOutExpo',
                    'do': function (v, t) {
                        popup.position.x = v;
                    },
                    onComplete: function () {
                        popup.animate();
                    }
                });

                // move out critter cards
                new Tween({
                    from: settings.grid.parent.position.y,
                    to: settings.grid.parent.position.y + 128,
                    'in': 40,
                    ease: 'exponential',
                    alpha: -6,
                    beta: 1,
                    'do': function (v, t) {
                        settings.grid.parent.position.y = v;
                    }
                });
            }
        });
        /**
         * Drop coins at "other" entity
         */
        var dropCoins = function (other, amount) {
            var delay = 0;

            var amountRemaining = amount;
            var dropList = [];

            var ii;
            for (ii = 0; ii < Utils.coinSettings.length; ii++) {
                var coinSettings = Utils.coinSettings[ii];
                while (amountRemaining >= coinSettings.amount) {
                    dropList.push(coinSettings);
                    amountRemaining -= coinSettings.amount;
                }
            }
            var currentDropIndex = 0;


            var spawnCoin = function () {
                // drop coins
                var coin;
                var currentCoin = dropList[currentDropIndex++];

                coin = new Coin({
                    position: other.position.add(new Vector2(0, -other.dimension.height / 2)),
                    speed: new Vector2(Utils.getRandomRangeFloat(-1, 0), Utils.getRandomRangeFloat(-1, -4)),
                    manager: manager,
                    amount: currentCoin.amount,
                    animationKey: currentCoin.animationKey,
                    scale: currentCoin.scale
                });
                coin.setDelay(delay);
                delay += 2;
                Bento.objects.attach(coin);
            };

            // Utils.repeat(Utils.getRandom(amount), spawnCoin);
            Utils.repeat(dropList.length, spawnCoin);
            Utils.trackCoins(amount);
        };
        /**
         * Check if one type is stronger than the other
         */
        var compareElements = function (element1, element2) {
            var strengths = {
                'water': 'fire',
                'fire': 'air',
                'earth': 'water',
                'air': 'earth'
            };
            var weakness = {
                'water': 'earth',
                'fire': 'water',
                'earth': 'air',
                'air': 'fire'
            };
            if (!element1 || !element2) {
                return 1;
            }

            if (strengths[element1] === element2) {
                return 2;
            } else if (weakness[element1] === element2) {
                return 0.5;
            } else {
                // unknown
                return 1;
            }
        };
        /**
         * Enemy running away animation
         */
        var runAway = function () {
            // enemy runs away
            frontEnemy.getComponent('walk', function (walkComponent) {
                walkComponent.getUpAndRun({
                    hasEscaped: function () {
                        victory({
                            escaped: true
                        });
                    }
                });
            });
        };
        /**
         * Caught enemy successfully
         */
        var catchEnemy = function () {
            var enemy = Utils.clone(frontEnemy.card.getProperties());
            var critters = Utils.getCritters();
            var team = Utils.getTeam();
            Bento.objects.remove(frontEnemy);

            // boss: swap with normal boss card
            if (isBossBattle) {
                enemy = Utils.clone(Bento.assets.getJson('creatures/creatures').creatures[enemy.id]);
                // Achievement: Collect a boss critter
                Utils.postAchievement('CgkI--zjoo8UEAIQBw', function () {});
            }
            // mechaboss: swap with normal mechaboss card
            else if (enemy.id.indexOf('mecha') > -1) {
                enemy = Utils.clone(Bento.assets.getJson('creatures/creatures').creatures[enemy.id]);
                Utils.worldMap.collectedMecha[enemy.id] = true;
            }

            enemy.currentHp = enemy.hp;

            // put in team right away if possible (after tutorial)
            if (Bento.saveState.load('sawTutorial2', false) === true) {
                if (team[0] === null) {
                    team[0] = enemy;
                } else if (team[1] === null) {
                    team[1] = enemy;
                } else if (team[2] === null) {
                    team[2] = enemy;
                } else {
                    critters.unshift(enemy);
                }
            } else {
                critters.unshift(enemy);
            }
            // does player have this critter already?
            newCritter = !Utils.hasInCollection(enemy.id);
            Utils.addToCollection(enemy.id);

            Utils.saveCritters();
            Utils.saveTeam();
        };
        /**
         * Start animation for catching
         */
        var extraChance = 0;
        var catchSequence = function (doRetry) {
            // roll the dice
            var chance = frontEnemy.getProperty('catchChance') + extraChance;
            var dice = Utils.getRandom(100);
            var success = (dice <= chance);
            var isPlayingSound = false;
            var sound = 'sfx_spaceship_beam';
            var playSound = function () {
                if (!isPlayingSound) {
                    isPlayingSound = true;
                    Bento.audio.playSound(sound, true);
                }
            };
            var stopSound = function () {
                isPlayingSound = false;
                Bento.audio.stopSound(sound);
            };

            var showSpaceShip = function () {
                var spaceShip = new Entity({
                    z: Utils.Layers.SKY + 10,
                    name: 'spaceShip',
                    position: new Vector2(viewport.x + viewport.width / 2, viewport.y - 75),
                    originRelative: new Vector2(0.5, 1),
                    components: [
                        new Sprite({
                            imageName: 'worldmap/ship',
                        })
                    ]
                });
                var beamSprite = new Sprite(Bento.assets.getJson('animations/catchbeam'));
                var beam = new Entity({
                    z: Utils.Layers.BACKGROUND,
                    name: 'beam',
                    position: new Vector2(viewport.width / 2, 0),
                    originRelative: new Vector2(0.5, 0),
                    components: [beamSprite]
                });
                var showBeam = function () {
                    beam.alpha = 0.7;
                    beamSprite.setAnimation('start');
                    beam.position.y = spaceShip.position.y - 4;
                    Bento.objects.attach(beam);
                };
                var beamUp = function () {
                    // beam up
                    playSound();
                    frontEnemy.getComponent('walk', function (walkComponent) {
                        walkComponent.beamUp({
                            success: success,
                            target: spaceShip.position.y - 37,
                            onComplete: function () {
                                var card;
                                stopSound();
                                beamSprite.setAnimation('end', function () {
                                    Bento.objects.remove(beam);
                                    moveOut();
                                });

                                if (!success) {
                                    // try again?
                                    // for now: just run away
                                    // Utils.timeout(30, runAway);
                                    doRetry();
                                } else {
                                    // success popup
                                    // showCaughtPopup();

                                    // twitter battle
                                    if (isTwitterBattle) {
                                        // heal
                                        card = Utils.clone(frontEnemy.card.getProperties());
                                        card.currentHp = card.hp;
                                        Utils.addToTwitterCritters(card);
                                    } else {
                                        catchEnemy();
                                    }
                                    // pass along if caught critter is a new one
                                    frontEnemy.newCritter = newCritter;
                                    victory({
                                        caught: frontEnemy
                                    });
                                }
                            }
                        });
                    });
                };
                var moveIn = function () {
                    // move in
                    new Tween({
                        from: viewport.y - 75,
                        to: viewport.y + 64,
                        'in': 60,
                        ease: 'squareroot',
                        'do': function (v, t) {
                            spaceShip.position.y = v;
                        },
                        onComplete: function () {
                            // show beam
                            showBeam();
                            beamUp();
                        }
                    });
                };
                var moveOut = function () {
                    new Tween({
                        from: spaceShip.position.y,
                        to: spaceShip.position.y - 80,
                        'in': 60,
                        ease: 'easeInBack',
                        'do': function (v, t) {
                            spaceShip.position.y = v;
                        },
                        onComplete: function () {
                            Bento.objects.remove(spaceShip);
                        }
                    });
                };

                moveIn();

                Bento.objects.attach(spaceShip);
            };

            showSpaceShip();
        };
        /**
         * Deprecated
         */
        var showCaughtPopup = function () {
            new CaughtPopup({
                frontEnemy: frontEnemy
            });
        };
        var groundY = viewport.height * 2 / 3 - 17;
        /*
         * Battle is over, player is victorious (may or may not have caught critter)
         * @param {Object} settings
         * @param {Entity} settings.caught - enemy caught
         */
        var victory = function (settings) {
            var caught = settings.caught;
            var escaped = settings.escaped;
            var alreadyDefeated = true;
            var dialog;
            var dialogText = [];
            var attachments = {};
            var item;
            var coins;
            var currentPlanet;
            // minutes it took until defeating the boss for the first time
            var minutes;
            // amount of tries it took to beat the boss
            var bossTries;
            var amountOfTwitterWins;
            var planets = ['starti', 'veggi', 'zaspiri', 'dark', 'hotmountain', 'candy', 'pirate', 'lavalli', 'mechastation'];
            var i = 0;

            // defeated boss for the first time?
            if (isBossBattle && !isTwitterBattle) {
                alreadyDefeated = Utils.hasDefeatedBoss(boss.id);
                if (!alreadyDefeated) {
                    Utils.defeatedBoss(boss.id);
                    // show popup and then leave to planet map

                    // boss always drops an upgrade for combo machine
                    dialogText.push(
                        (Utils.Localization.getText('dropped')).replace('{NAME}', Utils.Localization.getName(frontEnemy.card.getProperty('id')))
                    );
                    dialogText.push(Utils.Localization.getText('upgrade'));
                    attachments['1'] = new Entity({
                        name: 'upgrade',
                        position: new Vector2(0, 12),
                        originRelative: new Vector2(0.5, 0.5),
                        components: [
                            new Sprite({
                                image: Bento.assets.getImage('battle/upgrade')
                            })
                        ]
                    });
                    // unlock rarity
                    Bento.saveState.save('unlockedRarity', Bento.saveState.load('unlockedRarity', 3) + 1);
                    dialogText.push(Utils.Localization.getText('unlockedRarity').replace("{RANK}", Bento.saveState.load('unlockedRarity', 3)));

                    currentPlanet = Bento.saveState.load('currentPlanet', null);
                    minutes = Bento.saveState.load('lastSentTimeThreshold_' + currentPlanet, -1);
                    bossTries = Bento.saveState.load('' + currentPlanet + 'BossAttempt', 0);
                    // how long it took to beat the planet, unique event per amount of minutes
                    Utils.Analytics.event('Planets', '' + currentPlanet + 'FnshdIn_' + minutes);
                    // how long it took to beat the planet, one event for everybody (to get an average)
                    Utils.Analytics.event('Planets', '' + currentPlanet + 'FnshdInMins', minutes);
                    // how many tries it took to beat the boss, unique event per amount of tries
                    Utils.Analytics.event('Planets', '' + currentPlanet + 'BossDftdAftr_' + bossTries);
                    // how many tries it took to beat the boss, one event for everybody (to get an average)
                    Utils.Analytics.event('Planets', '' + currentPlanet + 'BossDftdAft', bossTries);
                    // save that boss on this planet is defeated
                    Bento.saveState.save('' + currentPlanet + 'BossDefeated', true);

                    Utils.Firebase.logEvent('finished_planet', {
                        planet: currentPlanet,
                        minutes_to_finish: minutes,
                        boss_tries: bossTries
                    });
                    // which level/planet has player finished. Standard Firebase event
                    for (i = 0; i < planets.length; ++i) {
                        if (planets[i] === currentPlanet) {
                            Utils.Firebase.logEvent(Utils.Firebase.Events.LEVEL_UP, {
                                level: i + 1
                            });
                            break;
                        }
                    }

                    if (Utils.worldMap.planetData.reward === 'map') {
                        dialogText.push(
                            (Utils.Localization.getText('galaxyMap')).replace('{NAME}', Utils.Localization.getName(frontEnemy.card.getProperty('id')))
                        );
                        attachments['3'] = new Entity({
                            name: 'mapPiece',
                            position: new Vector2(-2, 16),
                            originRelative: new Vector2(0.5, 0.5),
                            components: [
                                new Sprite({
                                    imageName: Utils.worldMap.planetData.rewardImage
                                })
                            ]
                        });
                        dialogText.push(Utils.Localization.getText('explore'));
                        console.log('unlocks', Utils.worldMap.planetData.unlocks);
                        // we save this value so any time you start the game, the map screen will unlock a planet
                        Bento.saveState.save('planetsToUnlock', Utils.worldMap.planetData.unlocks || []);
                        // track GPG world completion events: https://play.google.com/apps/publish/?dev_acc=14595537485418246873#EventsPlace:gt=691294238331
                        trackCompletion(currentPlanet);
                    }

                    dialog = new StoryDialog({
                        text: dialogText,
                        attachments: attachments,
                        onClose: function () {
                            new Fade({
                                callback: function () {
                                    Bento.screens.show('screens/map', {
                                        battled: true,
                                        caught: caught,
                                        // unlocked: Utils.worldMap.planetData.reward === 'map',
                                        showTransition: !alreadyDefeated
                                    });
                                }
                            });
                        }
                    });

                    return;
                } else {
                    // give a reward?
                }
            }

            // give a reward?
            // coins = (frontEnemy.card.getProperty('rarity') || 1) * 2 + Utils.getRandom(10);
            // dialog = new Dialog({
            //     text: ("\n{name} dropped " + coins + " coins!").replace('{name}', frontEnemy.card.getProperty('name')),
            //     type: 'yes',
            //     canClose: false,
            //     onClose: function () {
            //         var myCoins = Bento.saveState.load('myCoins', 0);
            //         myCoins += coins;
            //         Bento.saveState.save('myCoins', myCoins);

            //         new Fade({
            //             callback: function () {
            //                 Bento.screens.show('screens/worldmap', {
            //                     battled: true,
            //                     caught: caught
            //                 });
            //             }
            //         });
            //     }
            // });
            Utils.timeout(60, function () {
                new Fade({
                    callback: function () {
                        if (isTwitterBattle) {
                            Bento.screens.show('screens/twitter', {
                                battled: true,
                                caught: caught,
                                escaped: escaped
                            });
                        } else {
                            Bento.screens.show('screens/worldmap', {
                                battled: true,
                                caught: caught
                            });
                        }
                    }
                });
            });

            if (isTwitterBattle) {
                amountOfTwitterWins = Bento.saveState.load('amountOfTwitterWins', 0);
                ++amountOfTwitterWins;
                Bento.saveState.save('amountOfTwitterWins', amountOfTwitterWins);
                Utils.Analytics.event('Twitter', 'wins_' + amountOfTwitterWins);
                // total twitter battles. Divide value by amount of events for average
                Utils.Firebase.logEvent('twitter_battle', {
                    value: 1,
                    wins: amountOfTwitterWins,
                    losses: Bento.saveState.load('amountOfTwitterLosses', 0)
                });
            }
        };
        /*
         * Player lost
         */
        var gameOver = function () {
            var amountOfTwitterLosses;
            manager.state = 'defeat';

            Utils.timeout(60, function () {
                if (!isTwitterBattle) {
                    Utils.saveTeam();
                }
                new GameOver({
                    name: 'defeat',
                    isTwitterBattle: isTwitterBattle,
                    isBossBattle: isBossBattle,
                    onClosed: function () {
                        new Fade({
                            callback: function () {
                                if (isTwitterBattle) {
                                    Bento.screens.show('screens/twitter', {
                                        battled: true,
                                        opponent: frontEnemy
                                    });
                                } else {
                                    // go back to worldmap
                                    Bento.screens.show('screens/worldmap', {
                                        battled: true,
                                        wasDefeated: true
                                    });
                                }
                            }
                        });
                    }
                });
                Bento.objects.get('battlePauseButton', function (battlePauseButton) {
                    Bento.objects.remove(battlePauseButton);
                });
            });

            if (isTwitterBattle) {
                amountOfTwitterLosses = Bento.saveState.load('amountOfTwitterLosses', 0);
                ++amountOfTwitterLosses;
                Bento.saveState.save('amountOfTwitterLosses', amountOfTwitterLosses);
                Utils.Analytics.event('Twitter', 'loss_' + amountOfTwitterLosses);

                // total twitter battles. Divide value by amount of events for average
                Utils.Firebase.logEvent('twitter_battle', {
                    value: 1,
                    wins: Bento.saveState.load('amountOfTwitterWins', 0),
                    losses: amountOfTwitterLosses
                });
            }

            // Achievement: Lose a battle
            Utils.postAchievement('CgkI--zjoo8UEAIQCw', function () {});
        };
        /**
         * Adds clickable behavior to cards
         */
        var addClickable = function (card) {
            var special = card.getProperty('specialAttack');
            var combo = card.getProperty('combo');
            var attackButton = new AttackButton({
                coolDown: card.getProperty('coolDown'),
                onClick: function () {
                    // start minigame
                    var doMinigame = function () {
                        console.log('Start minigame');
                        // Minigame = Utils.pickRandom(Swipes, Catch, Trace, Spin, Order);

                        currentSpecialAttack = new MiniGame({
                            rarity: frontCritter.card.getProperty('rarity'),
                            critter: frontCritter,
                            manager: manager,
                            enemy: frontEnemy,
                            attack: attack,
                            combo: combo,
                            onHit: function () {
                                // TODO: combo effects
                            },
                            onComplete: function (success, max) {
                                Bento.objects.remove(currentSpecialAttack);
                                currentSpecialAttack = null;
                            }
                        });
                        Bento.objects.attach(currentSpecialAttack);
                    };
                    // if (frontCritter.getComponent('walk').isIdling()) {
                    //     doMinigame();
                    // }
                    doMinigame();
                    // TODO: every other card/buttons should be deactivated
                    // what to do if the critter to be attacked is about to die (by an incoming attack from another critter?)
                    // cancel all buttons until attacks are finished...?
                }
            });
            var attack = function (settings) {
                var i;
                var damageBoost = settings ? settings.damageBoost || 1 : 1;
                var lastCombo;

                if (!frontEnemy) {
                    // nobody to attack?
                    console.log('Attack canceled: nobody to attack');
                    return;
                }
                // is critter alive?
                if (card.critter.getProperty('currentHp') <= 0) {
                    console.log('Attack canceled: critter is dead');
                    return;
                }
                // game is over
                if (manager.state !== 'battle') {
                    console.log('Attack canceled: battle is over');
                    return;
                }
                // already attacking
                if (isAnyAttacking) {
                    console.log('Attack canceled: was already attacking');
                    return;
                }

                // this critter becomes the front critter
                frontCritter = card.critter;

                // other critters should walk away
                for (i = 0; i < critters.length; ++i) {
                    if (critters[i] !== card.critter) {
                        critters[i].getComponent('walk', function (component) {
                            component.targetX = -64;
                        });
                    }
                }

                //remove tutorial hand if visible
                if (Bento.saveState.load('sawTutorial', false) === false) {
                    for (i = 0; i < card.components.length; ++i) {
                        if (card.components[i].name === "hand") {
                            card.components[i].visible = false;
                        }
                    }
                }

                // attack!!
                card.getAvatar().alpha = 0.5;
                card.critter.getComponent('walk', function (walkComponent) {
                    // TODO: make other kinds of attack, other than jumpAttack
                    walkComponent.jumpAttack({
                        targetX: frontEnemy.position.x - 32,
                        other: frontEnemy,
                        onHit: function (self, other) {
                            var hasKilled = false;
                            manager.onHit(self, other, damageBoost, function () {
                                // cool slowmo if the critter died from special attack
                                if (currentSpecialAttack) {
                                    Utils.screenShake(3, 5);
                                    new Tween({
                                        'in': 60,
                                        ease: 'linear',
                                        updateWhenPaused: Utils.PauseLevels.ALWAYS,
                                        ignoreGameSpeed: true,
                                        'do': function (v, t) {
                                            Bento.setGameSpeed(0.2);
                                        },
                                        onComplete: function () {
                                            Bento.setGameSpeed(1);
                                        }
                                    });

                                    // finish special attack early if it was the last enemy
                                    if (frontEnemy === enemies[enemies.length - 1]) {
                                        hasKilled = true;
                                        Bento.objects.remove(currentSpecialAttack);
                                        currentSpecialAttack = null;
                                    }
                                }
                            });

                            // do combo?
                            if (!hasKilled && settings && settings.onHit) {
                                lastCombo = settings.onHit();
                                // keep track of how many special attacks were completed in a row for achievement
                                if (lastCombo) {
                                    completedSpecialAttacksInARow++;
                                }

                                // Achievement: In a single battle perform 3 special attacks in a row without any normal attacks in between
                                if (completedSpecialAttacksInARow === 3) {
                                    Utils.postAchievement('CgkI--zjoo8UEAIQCQ', function () {});
                                }
                            } else {
                                Bento.audio.stopSound('sfx_battle_bump');
                                Bento.audio.playSound('sfx_battle_bump');

                                // keep track of how many special attacks were completed in a row for achievement
                                completedSpecialAttacksInARow = 0;
                            }
                        },
                        onFinished: function (self, other) {
                            var enemyBehavior = other.getComponent('enemyBehavior');
                            // this is where the turn is over
                            currentTurn += 1;
                            console.log('Turn is over');

                            // set enemy turn to see if they can attack
                            if (enemyBehavior) {
                                enemyBehavior.turn -= 1;
                                if (enemyBehavior.turn < 0) {
                                    enemyBehavior.turn = 0;
                                }
                            }

                            // set player turn for special attack
                            // Utils.forEach(critters, function (critter, i, l, breakLoop) {
                            //     if (critter) {
                            //         critter.card.getComponent('attackButton', function (abilityButton) {
                            //             abilityButton.addTurn();
                            //         });
                            //     }
                            // });

                            //
                            if (card.critter.getProperty('currentHp') <= 0) {
                                return;
                            }
                            card.getAvatar().alpha = 1;

                            card.getComponent('attackButton', function (abilityButton) {
                                abilityButton.addTurn();
                            });
                        }
                    });

                    // walk to front
                    walkComponent.targetX = 32;
                });
            };

            // load minigame
            if (special) {
                // bento.require(['entities/battle/minigames/catch'/* + special*/], function (module) {
                //     Minigame = module;
                // });

                // disable self if any critter is in the middle of an attack
                attackButton.attach({
                    name: 'checkAttacking',
                    isActive: true,
                    update: function () {
                        attackButton.setActive(!isAnyAttacking &&
                            !currentSpecialAttack &&
                            card.critter.getProperty('currentHp') > 0 &&
                            manager.state === 'battle' &&
                            attackButton.visible
                        );
                    }
                });
                card.attach(attackButton.getBubbles());
                card.attach(attackButton);
            }

            // normal attack
            card.attach(new Clickable({
                onClick: function () {
                    if (frontCritter && frontCritter.getComponent('walk').isIdling() && !currentSpecialAttack && !isEnemyAttacking) {
                        attack();
                    }
                }
            }));
        };
        /**
         * Create card entities and battle entities from array of critters settings
         */
        var loadCritters = function (critters) {
            var i = 0;
            var card;
            var width = 0;
            var height = 0;
            var critter;
            var abilityButton;
            for (i = 0; i < critters.length; ++i) {
                if (critters[i] === null) {
                    settings.grid.addCard(null);
                    continue;
                }
                card = new Card(critters[i]);
                width = card.dimension.width + 2;
                height = card.dimension.height;
                addClickable(card);
                // add to grid
                settings.grid.addCard(card);

                //show a hand during first battle on the first card only
                if (i === 0 && Bento.saveState.load('sawTutorial', false) === false) {
                    card.attach(new Entity({
                        name: 'hand',
                        position: new Vector2(0, 8),
                        originRelative: new Vector2(0.5, 0),
                        updateWhenPaused: Utils.PauseLevels.CAUGHT,
                        components: [new Sprite({
                            imageName: 'map/hand'
                        })]
                    }).attach({
                        update: function (data) {
                            var hand = data.entity;
                            hand.position.y = 18 + 4 * Math.abs(Math.sin(hand.timer / 20));
                        }
                    }));
                }

                // make corresponding critter entity
                card.critter = makeCritter(critters[i]);
                card.critter.card = card;
                // card.abilityButton = abilityButton;
                card.visible = false;

                // critter is already dead
                if (card.getProperty('currentHp') <= 0) {
                    card.getAvatar().alpha = 0.25;
                }
            }
            if (!frontCritter) {
                // uh oh, all critters are dead already?!
                gameOver();
            }
        };
        /**
         * Instantiate critter with critter settings
         */
        var makeCritter = function (properties) {
            var walkComponent = new Walk({});
            var critter = new Critter(properties);
            critter.position = new Vector2(-64, groundY);
            critter.attach(walkComponent);
            Bento.objects.attach(critter);

            walkComponent.targetX = critter.position.x;

            critters.push(critter);
            // let the first critter walk forward
            if (!frontCritter) {
                if (properties.currentHp <= 0) {
                    // the critter is dead?
                    // lets wait for the next critter
                } else {
                    frontCritter = critter;
                    walkComponent.targetX = 32;
                }
            }

            return critter;
        };
        /**
         * Instantiate enemy critter with critter properties
         */
        var makeEnemy = function (properties, totalEnemies) {
            // TEMP
            var enemy = new Critter(properties);
            var enemyComponent = new Enemy(properties, manager);
            var walkComponent = new Walk({
                defaultScaleX: -1,
                isTwitter: isTwitterBattle
            });
            enemy.timer = 30;
            enemies.push(enemy);
            Bento.objects.attach(enemy);

            // enemy.position = new Vector2(viewport.width + 32, groundY);
            enemy.position = new Vector2(viewport.width / (totalEnemies + 1) * enemies.length, groundY);
            enemy.attach(walkComponent);
            enemy.attach(enemyComponent);
            walkComponent.targetX = enemy.position.x;

            // don't attack yet
            // enemyComponent.turn = 1;
            if (isTwitterBattle) {
                enemyComponent.turn = enemy.getProperty('turnsToAttack');
            } else {
                enemyComponent.turn = 1;
            }
            // let the first critter walk forward
            // if (enemies.length === 1) {
            //     frontEnemy = enemy;
            //     // don't attack yet
            //     enemyComponent.turn = 1;
            //     // walk forward
            //     walkComponent.targetX = viewport.width - 32;
            // } else {
            //     // walk away
            //     walkComponent.targetX = viewport.width + 32;
            // }

            return enemy;
        };
        var makeEnemies = function () {
            var i;
            var card;
            var width = 0;
            var height = 0;
            var enemy;

            // sort by rarity (?)
            enemyCritters.sort(function (a, b) {
                return a.rarity - b.rarity;
            });

            // make enemies
            for (i = 0; i < enemyCritters.length; ++i) {
                // boss enemy? Give the boss extra health
                if (isBossBattle) {
                    enemyCritters[i].hp = settings.data.boss.hp || enemyCritters[i].hp;
                    enemyCritters[i].attack = settings.data.boss.attack || enemyCritters[i].attack;
                    enemyCritters[i].currentHp = enemyCritters[i].hp;
                }

                if (isTwitterBattle) {
                    card = new Card(enemyCritters[i], {
                        image: enemyCritters[i].image
                    });
                } else {
                    card = new Card(enemyCritters[i]);
                    card.getAvatar().scale.x = -1;
                }
                width = card.dimension.width + 2;
                height = card.dimension.height;

                // add to grid
                settings.enemyGrid.addCard(card);

                // make corresponding critter entity
                enemy = makeEnemy(enemyCritters[i], enemyCritters.length);
                card.critter = enemy;
                card.critter.card = card;

                card.visible = false;
                // show card
                (function (enemyCard, t) {
                    Utils.timeout(10 * t, function () {
                        enemyCard.visible = true;
                        showCard(enemyCard, 'move');
                    });
                })(card, i);
            }

        };
        /**
         * Intro for normal battle
         */
        var introNormal = function () {
            var exclamation = new Entity({
                z: Utils.Layers.CRITTERS,
                name: 'exclamation',
                position: new Vector2(viewport.width / 2, groundY - 64),
                originRelative: new Vector2(0.5, 0.5),
                components: [
                    new Sprite({
                        imageName: 'worldmap/attackbubble'
                    })
                ]
            });
            var bossTries;
            var currentPlanet = Bento.saveState.load('currentPlanet', null);

            Bento.objects.attach(exclamation);
            Bento.audio.playSound('sfx_jump');
            new Tween({
                from: 0,
                to: 1,
                'in': 30,
                ease: 'elastic',
                alpha: -5,
                beta: 2,
                'do': function (v, t) {
                    exclamation.scale.x = 2 - v;
                    exclamation.scale.y = v;
                },
                onComplete: function () {
                    Utils.timeout(30, function () {
                        Bento.objects.remove(exclamation);
                    });
                }
            });

            // enemies get scared
            Utils.forEach(enemies, function (enemy, i, l) {
                Utils.timeout(10 * i, function () {
                    enemy.getComponent('walk', function (component) {
                        component.jump(5);
                    });
                });
            });

            Utils.timeout(50, function () {
                // move enemies back
                Utils.forEach(enemies, function (enemy, i, l) {
                    var walk = enemy.getComponent('walk');
                    if (i === 0) {
                        walk.targetX = viewport.width - 32;
                        frontEnemy = enemy;
                    } else {
                        walk.targetX = viewport.width + 32;
                    }
                });
                // move players in
                startBattle();
            });

            // if boss not defeated yet, track how many tries it takes
            if (!isTwitterBattle && isBossBattle && !Bento.saveState.load('' + currentPlanet + 'BossDefeated', false)) {
                bossTries = Bento.saveState.load('' + currentPlanet + 'BossAttempt', 0);
                ++bossTries;
                Bento.saveState.save('' + currentPlanet + 'BossAttempt', bossTries);
                Utils.Analytics.event('Planets', '' + currentPlanet + 'BossAttmpt_' + bossTries);
                // boss attempts per planet. Divide value by amount of events for average
                Utils.Firebase.logEvent(currentPlanet + '_boss_attempt', {
                    value: 1,
                    attempt: bossTries
                });
            }

            // end of intro = start of battle
            if (!isTwitterBattle && isBossBattle && Bento.saveState.load('bossTutorial', false) === false) {
                new StoryDialog({
                    text: Utils.Localization.getText('bossTutorial'),
                    attachments: {
                        '1': new Entity({
                            name: 'upgrade',
                            position: new Vector2(0, 12),
                            originRelative: new Vector2(0.5, 0.5),
                            components: [
                                new Sprite({
                                    image: Bento.assets.getImage('battle/upgrade')
                                })
                            ]
                        })
                    },
                    onClose: function () {
                        Bento.saveState.save('bossTutorial', true);
                    }
                });

            }
        };
        var showCard = function (card, type) {
            type = type || 'move';
            if (type === 'twist') {
                new Tween({
                    from: 0,
                    to: 1,
                    'in': 60,
                    ease: 'cuberoot',
                    'do': function (v) {
                        var oscillations = 2;
                        var t = Math.PI * 2 * oscillations + Math.PI * 0.5;
                        card.scale.x = Math.sin(t * v);
                    },
                    onComplete: function () {
                        card.scale.x = 1;
                    }
                });
            } else if (type === 'move') {
                Bento.audio.stopSound('sfx_battle_card');
                Bento.audio.playSound('sfx_battle_card');
                new Tween({
                    from: card.position.x - viewport.width,
                    to: card.position.x,
                    'in': 60,
                    ease: 'elastic',
                    alpha: -10,
                    beta: 1,
                    'do': function (v) {
                        card.position.x = v;
                    },
                    onComplete: function () {
                        // TODO activate card
                    }
                });
            }
        };
        var startBattle = function () {
            var timeOfCurrentBattleStart = PlayTimer.playTimer;
            // time in 10 second increments
            var timeBetweenBattleStarts;
            // time in seconds
            var timeBetweenBattleStartsSeconds;
            var currentPlanet;
            var getHand = function () {
                return new Entity({
                    name: 'hand',
                    position: new Vector2(0, 8),
                    originRelative: new Vector2(0.5, 0),
                    components: [new Sprite({
                        imageName: 'map/hand'
                    })]
                }).attach({
                    update: function (data) {
                        var hand = data.entity;
                        hand.position.y = 4 + 4 * Math.abs(Math.sin(hand.timer / 20));
                    }
                });
            };
            var fleeTutorial = function () {
                var planetName = Utils.worldMap.planetData.planetName;
                if (isTwitterBattle) {
                    return;
                }
                if (planetName !== 'veggi') {
                    return;
                }
                if (Bento.saveState.load('sawFleeTutorial', false) === true) {
                    return;
                }

                new StoryDialog({
                    text: [
                        Utils.Localization.getText('fleeTutorial')
                    ],
                    attachments: {
                        "0": new Entity({
                            position: new Vector2(0, 20),
                            originRelative: new Vector2(0.5, 0.5),
                            components: [
                                new Sprite({
                                    image: Bento.assets.getImage('gui/tutorialflee')
                                })
                            ]
                        })
                    },
                    onClose: function () {
                        // show hand on flee button
                        Bento.objects.get('battlePauseButton', function (button) {
                            button.attach(getHand());
                        });

                        // save tutorial
                        Bento.saveState.save('sawFleeTutorial', true);
                    }
                });
            };

            // track time between starts of battles per planet
            battleStartTime = PlayTimer.playTimer;
            Bento.saveState.save('timeOfPrevBattleStart', PlayTimer.playTimer);
            currentPlanet = Bento.saveState.load('currentPlanet', null);
            timeBetweenBattleStarts = Math.floor((timeOfCurrentBattleStart - timeOfPrevBattleStart) / 60 / 10) * 10;
            timeBetweenBattleStartsSeconds = Math.floor((timeOfCurrentBattleStart - timeOfPrevBattleStart) / 60);
            Utils.Analytics.event('Planets', '' + currentPlanet + 'TimBtwnBttl_' + timeBetweenBattleStarts);
            Utils.Analytics.event('Planets', '' + currentPlanet + 'AvgTimeBtwnBttl', timeBetweenBattleStartsSeconds);
            // average time between battles per planet. Divide value by amount of events
            Utils.Firebase.logEvent(currentPlanet + '_time_tween_battles', {
                value: timeBetweenBattleStartsSeconds
            });

            // load your critters and cards
            loadCritters(playerCritters);
            settings.grid.parent.visible = true;
            // move in cards
            new Tween({
                from: settings.grid.parent.position.y + 64,
                to: settings.grid.parent.position.y,
                'in': 40,
                ease: 'exponential',
                alpha: -6,
                beta: 1,
                'do': function (v, t) {
                    settings.grid.parent.position.y = v;
                },
                onComplete: function () {
                    Bento.audio.playSound('sfx_battle_ready');
                    fleeTutorial();
                    Utils.forEach(critters, function (critter, i, l) {
                        if (!critter) {
                            return;
                        }
                        // show card
                        Utils.timeout(10 * i, function () {
                            critter.card.visible = true;
                            showCard(critter.card);
                        });
                    });
                    Utils.forEach(enemies, function (critter, i, l) {
                        if (!critter) {
                            return;
                        }
                        // get ready to fight
                        critter.getComponent('enemyBehavior', function (component) {
                            component.ready(true);
                        });
                        if (critter !== frontEnemy) {
                            new Tween({
                                from: 1,
                                to: 0,
                                'in': 15,
                                ease: 'easeInQuad',
                                updateWhenPaused: 0,
                                'do': function (v, t) {
                                    critter.alpha = v;
                                },
                                onComplete: function () {
                                    critter.visible = false;
                                    critter.alpha = 1;
                                }
                            });
                        }
                    });
                }
            });
        };
        // track duration of battles
        var sendBattleTime = function () {
            // battle time in 10's of seconds
            var battleTime = Math.floor((PlayTimer.playTimer - battleStartTime) / 60 / 10) * 10;
            var battleTimeSeconds = Math.floor((PlayTimer.playTimer - battleStartTime) / 60);
            var currentPlanet = Bento.saveState.load('currentPlanet', null);
            Utils.Analytics.event('Planets', '' + currentPlanet + 'BttlTime_' + battleTime);
            Utils.Analytics.event('Planets', '' + currentPlanet + 'AvgBttlTime', battleTimeSeconds);
            // average battle time per planet. Divide value by amount of events
            Utils.Firebase.logEvent(currentPlanet + '_battle_time', {
                value: battleTimeSeconds
            });
        };
        // track world completion with GPG events: https://play.google.com/apps/publish/?dev_acc=14595537485418246873#EventsPlace:gt=691294238331
        var trackCompletion = function (planet) {
            var planets = [
                'starti',
                'veggi',
                'zaspiri',
                'dark',
                'hotmountain',
                'candy',
                'pirate',
                'lavalli',
                'mechastation'
            ];
            var worldCompleted = [
                'CgkI--zjoo8UEAIQDw',
                'CgkI--zjoo8UEAIQEA',
                'CgkI--zjoo8UEAIQEQ',
                'CgkI--zjoo8UEAIQEg',
                'CgkI--zjoo8UEAIQEw',
                'CgkI--zjoo8UEAIQFA',
                'CgkI--zjoo8UEAIQFQ',
                'CgkI--zjoo8UEAIQFg',
                'CgkI--zjoo8UEAIQFw'
            ];
            var timeToCompletion = [
                'CgkI--zjoo8UEAIQGA',
                'CgkI--zjoo8UEAIQGQ',
                'CgkI--zjoo8UEAIQGg',
                'CgkI--zjoo8UEAIQGw',
                'CgkI--zjoo8UEAIQHA',
                'CgkI--zjoo8UEAIQHQ',
                'CgkI--zjoo8UEAIQHg',
                'CgkI--zjoo8UEAIQHw',
                'CgkI--zjoo8UEAIQIA'
            ];
            var i = 0;
            var il = 0;
            var timePlayed = 0;

            for (i = 0, il = planets.length; i < il; ++i) {
                if (planet === planets[i]) {
                    Utils.GooglePlay.submitEvent(worldCompleted[i], 1);
                    // amount of seconds spent in world before defeating boss
                    timePlayed = Math.floor(Bento.saveState.load('totalTimePlayedInWorld_' + planet, 0) / 60);
                    Utils.GooglePlay.submitEvent(timeToCompletion[i], timePlayed);
                }
            }

        };

        settings.grid.parent.visible = false;

        // load enemies
        makeEnemies();

        // check attacking
        manager.attach({
            name: 'checkAttacking',
            update: function () {
                var check = function (critter, i, l, breakLoop) {
                    if (!critter) {
                        return;
                    }
                    critter.getComponent('walk', function (walkComponent) {
                        if (walkComponent.isAttacking()) {
                            isAnyAttacking = true;
                            breakLoop();
                        }
                    });
                };
                var checkEnemy = function (critter, i, l, breakLoop) {
                    if (!critter) {
                        return;
                    }
                    critter.getComponent('walk', function (walkComponent) {
                        if (walkComponent.isAttacking()) {
                            isEnemyAttacking = true;
                            breakLoop();
                        }
                    });
                };
                isAnyAttacking = false;
                isEnemyAttacking = false;

                Utils.forEach(critters, check);
                Utils.forEach(enemies, checkEnemy);

            }
        });

        if (!isTwitterBattle && isBossBattle && Bento.saveState.load('firstBossBattle', false) === false) {
            Bento.saveState.save('firstBossBattle', true);
        }

        Zoom.init();

        // show abilities
        Utils.timeout(60, function () {
            // new AbilityBuy({
            //     onClosed: introNormal
            // });
            introNormal();
        });

        return manager;
    };
});