/**
 * Base of the minigames
 */
bento.define('entities/battle/minigames/base', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'entities/effect',
    'entities/battle/minigames/swipes',
    'entities/battle/minigames/catch',
    'entities/battle/minigames/trace',
    'entities/battle/minigames/spin',
    'entities/battle/minigames/order'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    ClickButton,
    Counter,
    Text,
    EventSystem,
    Utils,
    Tween,
    Effect,
    Swipes,
    Catch,
    Trace,
    Spin,
    Order
) {
    'use strict';
    return function (settings, title) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
            rarity: Number, critter rarity
            onHit: Function,
            onComplete: Function(Number success, Number max) // minigame is completed
            critter: frontCritter,
            manager: entity,
            enemy: frontEnemy,
            attack: Function
        }*/
        var viewport = Bento.getViewport();
        var currentTurnEntity = null;
        var maxTurns = settings.combo || 2;
        var turns = 0;
        var successTurns = 0;
        var SpeedComponent = function () {
            return {
                update: function (data) {
                    data.speed = 1;
                },
                attached: function (data) {
                    data.entity.moveComponentTo(this, 0);
                }
            };
        };
        /**
         * Text showing what to do
         */
        var text = new Text({
            position: new Vector2(0, -viewport.height / 3),
            text: title,
            font: 'volter',
            fontSize: 16,
            align: 'center',
            textBaseline: 'middle',
            fontColor: '#ffffff',
            strokeStyle: '#000000',
            maxWidth: viewport.width,
            lineWidth: 1,
            pixelStroke: true // workaround for Cocoon bug
        }).attach({
            name: 'textBehavior',
            timer: 0,
            update: function () {
                this.timer += 1;
                text.scale.x = 1 + 0.025 * Math.sin(this.timer / 20);
                text.scale.y = text.scale.x;
            }
        });
        /**
         * End of swipe
         * @parameter {Boolean} success - if the player did the swipe right
         */
        var endOfTurn = function (success) {
            // clean up
            entity.remove(overlay);
            entity.remove(text);
            entity.remove(currentTurnEntity);
            currentTurnEntity = null;

            if (!success) {
                onFail();
            } else {
                successTurns += 1;
            }

            // lighten overlay
            new Tween({
                from: 0.5,
                to: 0,
                'in': 30,
                ease: 'linear',
                'do': function (v, t) {
                    overlay.alpha = (v);
                },
                ignoreGameSpeed: true,
                updateWhenPaused: Utils.PauseLevels.MINIGAME,
                onComplete: function () {
                    // game over when max turns or missed a swipe
                    if (turns >= maxTurns || !success) {
                        if (settings.onComplete) {
                            settings.onComplete(successTurns, maxTurns);
                        }
                    }
                    // Bento.setGameSpeed(1);
                    Bento.objects.resume();

                    if (success) {
                        onSuccess();
                    }
                }
            });
        };
        var onSuccess = function () {
            // do a new attack!
            settings.attack({
                onHit: function () {
                    var effect;
                    var other;
                    var lastCombo;

                    // show combo effect
                    other = settings.manager.getFrontEnemy();
                    if (other) {
                        // Bento.objects.get('normalHit', function (normalHit) {
                        //     Bento.objects.remove(normalHit);
                        // });
                        effect = new Effect({
                            z: Utils.Layers.MINIGAME - 1,
                            sprite: new Sprite(Bento.assets.getJson('animations/effects/combo' + Math.min(turns, 3))),
                            position: other.position.add(new Vector2(0, -32)),
                            originRelative: new Vector2(0.5, 0.5)
                        });
                        effect.attach(new SpeedComponent());
                        Utils.screenShake(3 * Math.min(turns, 3), 5 * Math.min(turns, 3));

                        Bento.audio.playSound('sfx_battle_combo' + Math.min(turns, 3));
                    }

                    // can start a new swipe?
                    if (turns < maxTurns) {
                        startTurn();
                    } else {
                        lastCombo = true;
                    }

                    return lastCombo;
                },
                damageBoost: turns * 0.5
            });
        };
        /**
         * Player failed to do the minigame right
         */
        var onFail = function () {
            var wrongBehavior = {
                update: function () {
                    if (wrong.timer > 30) {
                        wrong.visible = Math.floor(wrong.timer / 2) % 2;
                    }
                    if (wrong.timer > 60) {
                        Bento.objects.remove(wrong);
                    }
                }
            };
            var wrong = new Entity({
                z: Utils.Layers.MINIGAME,
                name: 'wrong',
                position: new Vector2(viewport.width / 2, viewport.height / 2),
                originRelative: new Vector2(0.5, 0.5),
                components: [
                    new Sprite({
                        imageName: 'gui/wrong'
                    }),
                    wrongBehavior
                ]
            });
            Bento.objects.attach(wrong);

            // animate wrong entity
            new Tween({
                from: 0,
                to: 1,
                'in': 60,
                ease: 'elastic',
                alpha: -5,
                beta: 2,
                ignoreGameSpeed: true,
                updateWhenPaused: Utils.PauseLevels.MINIGAME,
                'do': function (v, t) {
                    wrong.scale.x = v;
                    wrong.scale.y = v;
                }
            });

            // Achievement: Mess up a special attack
            Utils.postAchievement('CgkI--zjoo8UEAIQCA', function () {});
        };
        var overlaySprite = new Sprite({
            imageName: 'fade'
        });
        var overlay = new Entity({
            name: 'overlay',
            position: new Vector2(0, 0),
            originRelative: new Vector2(0.5, 0.5),
            family: [''],
            components: [
                overlaySprite
            ]
        });
        var entity = new Entity({
            z: Utils.Layers.MINIGAME,
            name: 'attackMiniGame',
            position: new Vector2(viewport.width / 2, viewport.height / 2),
            family: [''],
            updateWhenPaused: Utils.PauseLevels.MINIGAME,
            components: [

            ]
        });
        var startTurn = function () {
            // fade in
            entity.attach(overlay);
            overlay.alpha = (0);
            new Tween({
                from: 0,
                to: 0.6,
                'in': 10,
                ease: 'linear',
                updateWhenPaused: Utils.PauseLevels.MINIGAME,
                ignoreGameSpeed: true,
                'do': function (v, t) {
                    overlay.alpha = (v);
                },
                onComplete: function () {
                    var game = Utils.pickRandom(Swipes, Catch, Trace, Spin, Order);
                    if (entity.isAdded && currentTurnEntity === null) {
                        // start
                        currentTurnEntity = game(settings)(endOfTurn, text);
                        turns += 1;
                        entity.attach(currentTurnEntity);
                    }
                    entity.attach(text);
                    // Bento.setGameSpeed(0.00001);
                    Bento.objects.pause(Utils.PauseLevels.MINIGAME);
                }
            });
        };

        // start with an attack, then proceed with the minigame
        settings.attack({
            onHit: function () {
                Bento.audio.stopSound('sfx_battle_bump');
                Bento.audio.playSound('sfx_battle_bump');
                startTurn();
            },
            damageBoost: 0.25
        });

        entity.extend({
            endOfTurn: endOfTurn
        });

        return entity;
    };
});