/**
 * Module description
 */
bento.define('entities/worldmap/chest', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/depth',
    'components/sort',
    'gui/storydialog'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    ClickButton,
    Counter,
    Text,
    EventSystem,
    Utils,
    Tween,
    Depth,
    Sort,
    StoryDialog
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var sprite = new Sprite({
            imageName: 'worldmap/chest',
            frameCountX: 2,
            animations: {
                default: {
                    frames: [0]
                },
                open: {
                    frames: [1]
                }
            }
        });
        var tiledId = settings.tiled.object.id;
        var reward = parseInt(settings.tiled.object.reward) || 10;
        var depth = new Depth({});
        var isOpened = false;
        var position = null;
        var planetName = Utils.worldMap.planetData.planetName;
        // 1 out of 50 chests is a mimic, they only spawn
        var isMimic = Bento.saveState.load('firstBossBattle', false) && (Utils.getRandom(15) === 0) && planetName !== "starti";
        // var isMimic = true;
        var behavior = {
            attached: function (data) {
                // mimic chests are triggers
                if (isMimic) {
                    data.entity.family.push('triggers');
                    data.entity.extend({
                        getHash: function () {
                            return 'mimic';
                        },
                        getEnemies: function () {
                            return [Utils.clone(Bento.assets.getJson('creatures/creatures').creatures['mimic'])];
                        }
                    });
                }
            },
            start: function () {
                var hash;
                // remember original position
                position = depth.getOriginalPosition() || entity.position.clone();

                // mimic chests are triggers
                if (isMimic) {
                    return;
                }

                // was already touched before
                if (Bento.saveState.load('chest-' + tiledId, false) === true) {
                    Bento.objects.remove(entity);
                } else {
                    // collided with player on start
                }
            },
            update: function () {
                var chestsFound;
                if (isOpened) {
                    entity.visible = entity.timer % 2;
                    if (entity.timer > 60) {
                        Bento.objects.remove(entity);
                    }
                    return;
                }
                if (isMimic) {
                    return;
                }
                entity.collidesWith({
                    family: 'player',
                    onCollide: function (other) {

                        // open chest
                        // Bento.objects.remove(entity);
                        other.stopSound();
                        sprite.setAnimation('open');
                        Bento.audio.playSound('sfx_combine');
                        new StoryDialog({
                            text: Utils.Localization.getText('chest').replace("{REWARD}", reward),
                            onClose: function () {
                                isOpened = true;
                                entity.timer = 0;

                                Bento.objects.get('myCoins', function (myCoins) {
                                    var coins = Bento.saveState.load('myCoins', 0);
                                    coins += reward;
                                    Utils.trackCoins(reward);
                                    Bento.saveState.save('myCoins', coins);
                                    EventSystem.fire('updateCoins');
                                    Bento.saveState.save('chest-' + tiledId, true);
                                });
                            }
                        });

                        chestsFound = Bento.saveState.load('chestsFound', 0);
                        // Achievement: Find 10 chests
                        if (++chestsFound >= 10)
                            Utils.postAchievement('CgkI--zjoo8UEAIQBg', function () {});
                        Bento.saveState.save('chestsFound', chestsFound);

                        // chest
                        Utils.repeat(10, spawnCoin);
                    }
                });
            }
        };
        var entity = new Entity({
            z: Utils.Layers.CRITTERS + 3,
            name: 'chest',
            position: settings.position || new Vector2(0, 0),
            originRelative: new Vector2(0.5, 1),
            boundingBox: new Rectangle(-8, -8, 16, 8),
            components: [
                sprite,
                depth,
                behavior,
                new Sort({})
            ]
        });
        var delay = 0;
        var spawnCoin = function () {
            var coinSprite = new Sprite(Bento.assets.getJson('animations/coinpickup'));
            var behavior = {
                lifeTime: Utils.getRandomRange(30, 40),
                speed: new Vector2(
                    Utils.getRandomRangeFloat(-1, 1),
                    Utils.getRandomRangeFloat(-4, -2)
                ),
                update: function () {
                    coin.position.x += this.speed.x;
                    coin.position.y += this.speed.y;

                    this.speed.y += 0.1;
                    if (coin.timer > this.lifeTime) {
                        coin.visible = coin.timer % 2;
                    }
                    if (coin.timer > this.lifeTime + 20) {
                        Bento.objects.remove(coin);
                    }
                }
            };
            var coin = new Entity({
                z: Utils.Layers.EFFECTS + 3,
                name: 'coin',
                position: entity.position.add(new Vector2(-8, -20)),
                originRelative: new Vector2(0, 0),
                updateWhenPaused: Utils.PauseLevels.STORY,
                components: [
                    coinSprite,
                    behavior
                ]
            });
            Utils.timeout(delay, function () {
                Bento.objects.attach(coin);
                Bento.audio.stopSound('sfx_coin');
                Bento.audio.playSound('sfx_coin');
            }, Utils.PauseLevels.STORY);
            delay += 5;
            return coin;
        };

        // entity.visible = false;
        return entity;
    };
});