/* globals THREE */
/**
 * 3D object
 * For cocoonjs, requires an edited three.js:
 * Inserted CanvasRenderer, remove _canvas.width = ... to stop it from resizing the canvas
 */
bento.define('entities/worldmap/mode7', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/packedimage'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    PackedImage
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var sprite;
        var entity;
        var size = new Rectangle(0, 0, Math.ceil(viewport.width), Math.ceil(viewport.height));

        var canvas = document.createElement('canvas');
        var originalContext = null;
        var cameraDistance = 0; // camera Y distance from player
        var meshes = [];
        var scene = new THREE.Scene();
        var camera = new THREE.PerspectiveCamera(60, size.width / size.height, 0.1, 1000);
        var canUseWebGl = function () {
            // try making a canvas
            try {
                var gl = canvas.getContext('webgl') || canvas.getContext('experimental-webgl');
                return !!window.WebGLRenderingContext;
            } catch (e) {
                return false;
            }
        };
        var useCanvas = Utils.isCocoonJs() || !canUseWebGl();
        var pixelSize = useCanvas ? 1 : 4;

        var ThreeRenderer = useCanvas ? THREE.CanvasRenderer : THREE.WebGLRenderer;
        // optimization for cocoon: render directly to screen
        var renderToScreen = useCanvas;

        var renderer = new ThreeRenderer({
            canvas: renderToScreen ? Bento.getCanvas() : canvas,
            // context: renderToScreen ? Bento.getContext() : undefined,
            // context: canvas.getContext('webgl'),
            alpha: true
        });
        // link with bento
        var threeRenderer = {
            name: '3DRenderer',
            start: function () {
                if (renderToScreen) {
                    EventSystem.on('preDraw', preDraw);
                    EventSystem.on('screenHidden', cleanEvents); // in case of bento refresh
                }
                entity.scale.x = 1 / pixelSize;
                entity.scale.y = entity.scale.x;
            },
            destroy: function () {
                if (renderToScreen) {
                    cleanEvents();
                }
            },
            // sets camera position
            update: function (data) {
                // calculate camera distance based on angle and angle
                cameraDistance = camera.position.z * Math.tan(-camera.rotation.x);
                camera.position.y = (viewport.y + cameraDistance);
                camera.position.x = -viewport.x;

            },
            draw: function (data) {
                if (!renderToScreen) {
                    canvas.texture = null;
                    renderer.render(scene, camera);
                }
            }
        };
        var debug = {
            press: 0,
            update: function () {
                // debugging
                // if (Bento.input.isKeyDown('7')) {
                //     this.press += 1;
                // } else {
                //     this.press = 0;
                // }
                // if (this.press === 1) {
                //     entity.visible = !entity.visible;
                // }

                // if (Bento.input.isKeyDown('9')) {
                //     camera.position.z += 1;
                //     console.log(camera.position.z);
                // }
                // if (Bento.input.isKeyDown('8')) {
                //     camera.position.z -= 1;
                //     console.log(camera.position.z);
                // }
                // if (Bento.input.isKeyDown('0')) {
                //     new Tween({
                //         from: camera.rotation.x,
                //         to: 0,
                //         'in': 180,
                //         ease: 'linear',
                //         'do': function (v, t) {
                //             camera.rotation.x = v;
                //         },
                //         onComplete: function () {

                //         }
                //     });

                // }
            }
        };
        var toBentoPos = function (vector) {
            vector.x = -vector.x - 48 + (size.width - 160) / 2; //???
            vector.y = vector.y + 4 - (size.height - 256) / 2;
            return vector;
        };
        var preDraw = function (data) {
            var originalContext = renderer.getContext();
            var bentoContext;

            if (!entity.visible) {
                return;
            }

            // swap context if needed (allows snapshots)
            if (data.renderer.getContext && renderToScreen) {
                bentoContext = data.renderer.getContext();

                // threejs resets imageSmoothing..???=
                if (bentoContext.imageSmoothingEnabled) {
                    bentoContext.imageSmoothingEnabled = false;
                }
                renderer.setContext(bentoContext);
            }
            // render now
            bentoContext.translate(Utils.screenSize.x, Utils.screenSize.y);
            renderer.render(scene, camera, function (ctx) {
                if (Utils.isWideScreen) {
                    // hack: need to offset the entire rendering process of three.js
                    ctx.translate(Utils.screenSize.x, (viewport.height - Utils.screenSize.height) * 2);
                }
            });

            // swap back
            if (data.renderer.getContext && renderToScreen) {
                renderer.setContext(originalContext);
            }
        };
        var cleanEvents = function () {
            EventSystem.off('preDraw', preDraw);
            EventSystem.off('screenHidden', cleanEvents);
        };

        renderer.autoClear = !useCanvas; // prevents holes
        if (renderToScreen) {
            if (Utils.isWideScreen) {
                // the camera viewport needs to be the original size!
                var multiplier = Bento.getCanvas().width / viewport.width;
                renderer.setSize(Bento.getCanvas().width, viewport.height * multiplier, false);
            } else {
                renderer.setSize(Bento.getCanvas().width, Bento.getCanvas().height, false);
            }
        } else {
            renderer.setSize(size.width * pixelSize, size.height * pixelSize);
        }

        camera.position.z = 150;
        camera.rotation.z = -Math.PI;
        camera.rotation.x = -Math.PI * 1 / 4;
        // camera.quaternion.setFromAxisAngle(new THREE.Vector3(1, 0, 0), Math.PI / 4);
        // shape.rotation.x = -Math.PI / 4;
        // shape.rotation.z = -Math.PI / 2;

        sprite = new Sprite({
            image: new PackedImage(canvas)
        });
        entity = new Entity({
            z: Utils.Layers.MODE7,
            name: 'mode7',
            position: new Vector2(viewport.width / 2, viewport.height / 2),
            originRelative: new Vector2(0.5, 0.5),
            updateWhenPaused: Utils.PauseLevels.CAUGHT,
            float: true,
            components: [
                threeRenderer,
                sprite,
                debug
            ]
        });

        entity.extend({
            makeSprite: function (position, image) {
                // Cocoon bug: image MUST be an Image element, not a canvas
                var texture;
                var material;
                var mesh;

                // cache texture
                if (image.texture) {
                    texture = image.texture;
                } else {
                    texture = new THREE.Texture(image);
                    texture.needsUpdate = true;
                    image.texture = texture;
                }
                material = new THREE.SpriteMaterial({
                    color: 0xffffff,
                    map: texture
                });
                mesh = new THREE.Sprite(material);

                // 
                position = toBentoPos(position);
                mesh.position.x = position.x;
                mesh.position.y = position.y;
                mesh.position.z = image.height / 2; // put it on its feet
                mesh.rotation.z = Math.PI;
                mesh.scale.x = 14; //???
                mesh.scale.y = 14;
                meshes.push(mesh);
                scene.add(mesh);
                return mesh;
            },
            makeMesh: function (position, image) {
                // Cocoon bug: image MUST be an Image element, not a canvas
                var geometry = new THREE.PlaneGeometry(image.width, image.height, 2, 2); // TODO use 1,2 for iphone 4s
                var texture;
                var material;
                var mesh;

                // cache texture
                if (image.texture) {
                    texture = image.texture;
                    material = image.material;
                    mesh = image.mesh.clone();
                } else {
                    texture = new THREE.Texture(image);
                    texture.needsUpdate = true;
                    texture.magFilter = THREE.NearestFilter;
                    texture.minFilter = THREE.NearestFilter;
                    material = new THREE.MeshBasicMaterial({
                        color: 0xffffff,
                        map: texture
                    });
                    mesh = new THREE.Mesh(geometry, material);

                    // save
                    image.texture = texture;
                    image.material = material;
                    image.mesh = mesh;
                }


                // 
                position = toBentoPos(position);
                mesh.position.x = position.x;
                mesh.position.y = position.y;
                mesh.rotation.z = Math.PI;
                meshes.push(mesh);
                scene.add(mesh);
                return mesh;

            },
            removeMesh: function (mesh) {
                var index = meshes.indexOf(mesh);
                scene.remove(mesh);
                if (index >= 0) {
                    meshes.splice(index, 1);
                }
            },
            worldToScreen: function (vec2) {
                var vec3, projScreenMat;
                var scale = vec2.add(new Vector2(32, 0));
                // converts 3D position to screen coords
                vec2 = toBentoPos(vec2);
                vec3 = new THREE.Vector3(vec2.x, vec2.y, 0);

                var v = vec3.project(camera);
                vec2.x = (v.x + 1) * size.width / 2;
                vec2.y = (-v.y + 1) * size.height / 2;
                // console.log(vec2)
                return new Vector2(viewport.x + vec2.x, viewport.y + vec2.y);
            },
            getCamera: function () {
                return camera;
            },
            getRenderer: function () {
                return renderer;
            }
        });

        return entity;
    };
});