/**
 * Player on world map
 */
bento.define('entities/worldmap/player', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/gui/clickbutton',
    'bento/gui/text',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/collidewithtrigger',
    'components/sort',
    'components/depth',
    'components/movable',
    'components/joystick'
], function (
    Bento,
    Vector2,
    Rectangle,
    ClickButton,
    Text,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    CollideWithTrigger,
    Sort,
    Depth,
    Movable,
    JoyStick
) {
    'use strict';
    var PI = Math.PI;
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
            planetData: Object //data coming from screens/map
        }*/
        var viewport = Bento.getViewport();
        var spriteShadow = new Sprite({
            imageName: 'worldmap/shadow'
        });
        var depth = new Depth({});
        var sort = new Sort({});
        var followPlayer = {
            name: 'followPlayer',
            update: function (data) {
                shadow.position.x = entity.position.x;
                shadow.position.y = entity.position.y;
                shadow.visible = entity.visible;
            }
        };
        var shadow = new Entity({
            z: Utils.Layers.CRITTERS + 1,
            name: 'shadow',
            position: new Vector2(0, 0),
            originRelative: new Vector2(0.5, 0.5),
            components: [
                spriteShadow,
                followPlayer
            ]
        });
        var sprite = new Sprite({
            imageName: 'worldmap/character',
            frameCountX: 4,
            frameCountY: 2,
            animations: {
                'up': {
                    speed: 0.03,
                    frames: [0, 4]
                },
                'down': {
                    speed: 0.03,
                    frames: [1]
                },
                'left': {
                    speed: 0.03,
                    frames: [2, 6]
                },
                'right': {
                    speed: 0.03,
                    frames: [3, 7]
                },
                'swirl': {
                    speed: 0.1,
                    frames: [0, 3, 1, 2]
                }
            }
        });
        var maxSpeed = 1.25;
        var movable = new Movable({
            maxSpeed: new Vector2(maxSpeed, maxSpeed)
        });
        var joystick = new JoyStick({});
        var isActivated = true;
        var controller = {
            tilt: 0,
            float: 0,
            start: function () {
                Bento.objects.attach(shadow);
            },
            destroy: function () {
                Bento.objects.remove(shadow);
            },
            update: function () {
                var angle;
                var speed;
                var animationName;
                
                if (!isActivated) {
                    return;
                }

                // keyboard controls
                if (Bento.input.isKeyDown('up')) {
                    movable.speed.y -= 0.5;
                } else if (Bento.input.isKeyDown('down')) {
                    movable.speed.y += 0.5;
                } else {
                    movable.speed.y = Utils.approach(movable.speed.y, 0, 0.2);
                }
                if (Bento.input.isKeyDown('left')) {
                    movable.speed.x -= 0.5;
                } else if (Bento.input.isKeyDown('right')) {
                    movable.speed.x += 0.5;
                } else {
                    movable.speed.x = Utils.approach(movable.speed.x, 0, 0.2);
                }

                // touch controls
                if (joystick.isPointerDown() && joystick.position.magnitude() > 0) {
                    movable.speed = joystick.position.scalarMultiply(1 / joystick.getRadius() * maxSpeed);
                    speed = movable.speed.magnitude();

                    this.float = Utils.approach(this.float, speed / maxSpeed, 1 / 30);

                    // animation
                    angle = movable.speed.angle();
                    if (Utils.isBetween(-PI / 4, angle, PI / 4)) {
                        sprite.setAnimation('right');
                        robo.rotation = this.float * Math.PI / 10;
                    }
                    if (Utils.isBetween(PI / 4, angle, PI * 3 / 4)) {
                        sprite.setAnimation('down');
                        robo.rotation = 0;
                    }
                    if (Utils.isBetween(-PI * 3 / 4, angle, -PI / 4)) {
                        sprite.setAnimation('up');
                        robo.rotation = 0;
                    }
                    if (Math.abs(angle) > PI * 3 / 4) {
                        sprite.setAnimation('left');
                    }

                    // shoot dust
                    if (robo.timer % 4 === 0) {
                        shootDustParticle(-angle);
                    }
                    // play sounds
                    playSound();
                } else {
                    this.float = Utils.approach(this.float, 0, 1 / 30);
                    stopSound();
                }

                animationName = sprite.getAnimationName();
                if (animationName === 'up' || animationName === 'down') {
                    robo.rotation = 0;
                } else if (animationName === 'left') {
                    robo.rotation = -this.float * Math.PI / 15;
                } else if (animationName === 'right') {
                    robo.rotation = this.float * Math.PI / 15;
                }

                // floating
                robo.position.y = (-2 + 2 * Math.sin(robo.timer / 20)) * this.float;
                shadow.scale.x = 0.9 + Math.abs(robo.position.y / 5) * 0.2;
                shadow.scale.y = shadow.scale.y;
            }
        };
        var moveCache = new Vector2(0, 0);
        var robo = new Entity({
            name: 'robo',
            position: new Vector2(0, 0),
            origin: new Vector2(16, 25),
            family: [''],
            components: [
                sprite
            ]
        });
        var entity = new Entity({
            z: Utils.Layers.CRITTERS + 3,
            name: 'player',
            position: settings.position || new Vector2(0, 0),
            boundingBox: new Rectangle(-8, -8, 16, 8),
            family: ['player'],
            components: [
                new Sprite({}), // legacy
                robo,
                movable,
                controller,
                new CollideWithTrigger({}),
                joystick
                // new Utils.Debug({}),
            ]
        });
        var shootDustParticle = function (angle) {
            var speed = 0.15;
            var behavior = {
                start: function () {
                    // angle += Utils.getRandomRangeFloat(-Math.PI / 2, Math.PI / 2);
                    angle += Utils.getRandomFloat(Math.PI * 2);
                    particle.getComponent('sprite', function (component) {
                        component.setAnimation('disappear', function () {
                            Bento.objects.remove(particle);
                        });
                    });
                },
                update: function () {
                    particle.position.x += -speed * Math.cos(angle);
                    particle.position.y += speed * Math.sin(angle);
                }
            };
            var particle = new Entity({
                z: Utils.Layers.CRITTERS + 1,
                name: 'dust',
                position: entity.position.add(new Vector2(-5, 0).rotateRadian(-angle)).add(new Vector2(0, -6)),
                originRelative: new Vector2(0.5, 0.5),
                components: [
                    new Sprite({
                        imageName: 'worldmap/wokje',
                        frameCountX: 2,
                        frameCountY: 2,
                        padding: 1,
                        animations: {
                            disappear: {
                                speed: 0.15,
                                frames: [0, 1, 2, 3]
                            }
                        }
                    }),
                    behavior
                ]
            });
            Bento.objects.attach(particle);
        };
        var isPlayingSound = false;
        var sound = 'sfx_player_move_loop';
        var playSound = function () {
            if (!isPlayingSound) {
                isPlayingSound = true;
                Bento.audio.playSound(sound, true);
            }
        };
        var stopSound = function () {
            isPlayingSound = false;
            Bento.audio.stopSound(sound);
        };
        sprite.setAnimation('down');

        //
        entity.extend({
            // allow player to move
            activate: function (bool) {
                isActivated = bool;
                joystick.activate(bool);
                if (!isActivated) {
                    movable.speed.x = 0;
                    movable.speed.y = 0;
                }
                // if (bool && !joystick.parent) {
                //     entity.attach(joystick);
                //     entity.attach(controller);
                // } else if (!bool && joystick.parent) {
                //     entity.remove(joystick);
                //     entity.remove(controller);
                // }
            },
            isActivated: function () {
                return isActivated;
                // return joystick.parent !== null;
            },
            setZ: function (height) {
                entity.transform.y = -height;
                shadow.alpha = (Utils.clamp(0, 1 - height / 64, 1));
            },
            addToWorld: function () {
                // entity.attach(depth);
                // entity.attach(sort);
            },
            removeFromWorld: function () {
                // entity.remove(depth);
                // entity.remove(sort);
                // entity.scale.x = 1;
                // entity.scale.y = 1;
            },
            setAnimation: function (name) {
                sprite.setAnimation(name);
            },
            releaseControls: function () {
                joystick.release();
            },
            stopSound: stopSound
        });

        return entity;
    };
});