/**
 * A dialog window with text and buttons
 * Automatically adds itself to game
 */
bento.define('gui/dialog', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/gui/clickbutton',
    'bento/gui/text'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    ClickButton,
    Text
) {
    'use strict';
    return function (settings) {
        /*settings = {
            name: name,
            canClose: Boolean, // default false
            type: String, // 'yesno' or 'yes'
            text: String or array, // string to display in window
            imageWindow: String, // imageName for dialog window image
            spriteYes: Sprite, // sprite for yes image, optional. Overrides settings.imagesYes
            spriteNo: Sprite, // sprite for no image, optional. Overrides settings.imagesNo
            imageYes: String, // imageName for yes image
            imageNo: String, // imageName for no image
            onYes: Function, // callback for pressing yes
            onNo: Function, // callback for pressing no
            onClose: Function // callback when closed, after pressing yes, no or close
            onNextPage: Function //
            pauseLevel: Number
        }*/
        var viewport = Bento.getViewport(),
            isMoving = false,
            type = settings.type || 'yesno',
            canClose = settings.canClose || false,
            buttonY = settings.buttonY || 50,
            spriteYes = settings.spriteYes || null,
            spriteNo = settings.spriteNo || null,
            imageYes = settings.imageYes || 'gui/dialog/buttonyes',
            imageNo = settings.imageNo || 'gui/dialog/buttonno',
            overlayOpacity = settings.overlayOpacity || 0.5,
            sprite = new Sprite({
                imageName: settings.imageWindow || 'gui/dialog/window'
            }),
            pauseLevel = settings.pauseLevel || Utils.PauseLevels.ALWAYS,
            page = 0,
            strings = settings.text || [],
            fade = new Entity({
                z: Utils.Layers.FADE - 2,
                name: 'fade',
                position: new Vector2(viewport.width / 2, viewport.height / 2),
                originRelative: new Vector2(0.5, 0.5),
                float: true,
                components: [
                    new Sprite({
                        imageName: 'fade'
                    })
                ]
            }),
            buttonClose = new ClickButton({
                sfx: 'sfx_button',
                image: Bento.assets.getImage('gui/buttonexit'),
                frameCountX: 2,
                frameCountY: 1,
                updateWhenPaused: pauseLevel,
                position: new Vector2(-74, -48),
                onClick: function () {
                    if (!isMoving) {
                        Utils.Analytics.event('Dialog', 'dialog-' + settings.name + '-close');
                        Utils.Firebase.logEvent('dialog_close', {
                            dialog_name: settings.name,
                            close_type: 'close'
                        });
                        moveOut();
                    }
                }
            }),
            addFade = {
                start: function () {
                    fade.alpha = 0;
                    Bento.objects.attach(fade);
                    new Tween({
                        from: 0,
                        to: overlayOpacity,
                        'in': 30,
                        ease: 'linear',
                        updateWhenPaused: pauseLevel,
                        'do': function (v, t) {
                            fade.alpha = v;
                        }
                    });
                },
                destroy: function () {
                    new Tween({
                        from: overlayOpacity,
                        to: 0,
                        'in': 10,
                        ease: 'linear',
                        updateWhenPaused: pauseLevel,
                        'do': function (v, t) {
                            fade.alpha = v;
                        },
                        onComplete: function () {
                            Bento.objects.remove(fade);
                        }
                    });
                }
            },
            entity = new Entity({
                z: Utils.Layers.FADE - 1,
                name: 'dialog',
                position: settings.position || new Vector2(viewport.x + viewport.width / 2, viewport.y + viewport.height / 4),
                originRelative: new Vector2(0.5, 0.5),
                updateWhenPaused: pauseLevel,
                components: [
                    sprite,
                    addFade
                ]
            }),
            text = new Text({
                z: 0,
                updateWhenPaused: pauseLevel,
                position: new Vector2(-6, -32),
                text: '', //settings.text || 'Dialog text',
                font: 'volter',
                fontSize: 9,
                fontSettings: Utils.FontSettings.get('font'),
                fontColor: '#ffffff',
                align: 'center',
                textBaseline: 'top',
                ySpacing: 1,
                maxWidth: 145,
                maxHeight: 60,
                // fontSettings: Utils.FontSettings.get('font')
            }),
            setupYesNoButtons = function () {
                var spriteYesButton = settings.spriteYes ||
                    new Sprite({
                        image: Bento.assets.getImage(imageYes),
                        frameCountX: 1,
                        frameCountY: 3,
                        animations: {
                            up: {
                                frames: [0]
                            },
                            down: {
                                frames: [1]
                            },
                            inactive: {
                                frames: [2]
                            }
                        }
                    });
                var spriteNoButton = settings.spriteNo ||
                    new Sprite({
                        image: Bento.assets.getImage(imageNo),
                        frameCountX: 1,
                        frameCountY: 3,
                        animations: {
                            up: {
                                frames: [0]
                            },
                            down: {
                                frames: [1]
                            },
                            inactive: {
                                frames: [2]
                            }
                        }
                    });
                var yesButton = new ClickButton({
                    name: 'buttonYes',
                    sfx: 'sfx_button',
                    sprite: spriteYesButton,
                    position: new Vector2(30, buttonY),
                    updateWhenPaused: pauseLevel,
                    onClick: function () {
                        if (isMoving) {
                            return;
                        }
                        if (settings.onYes) {
                            Utils.Analytics.event('Dialog', 'dialog-' + settings.name + '-yes');
                            Utils.Firebase.logEvent('dialog_close', {
                                dialog_name: settings.name,
                                close_type: 'yes'
                            });
                            settings.onYes();
                        }
                        moveOut();
                    }
                });
                var noButton = new ClickButton({
                    name: 'buttonNo',
                    sfx: 'sfx_button',
                    sprite: spriteNoButton,
                    position: new Vector2(-40, buttonY),
                    updateWhenPaused: pauseLevel,
                    onClick: function () {
                        if (isMoving) {
                            return;
                        }
                        if (settings.onNo) {
                            Utils.Analytics.event('Dialog', 'dialog-' + settings.name + '-no');
                            Utils.Firebase.logEvent('dialog_close', {
                                dialog_name: settings.name,
                                close_type: 'no'
                            });
                            settings.onNo();
                        }
                        moveOut();
                    }
                }).attach(
                    // back button pressed
                    new Utils.BackButton(function (data) {
                        data.entity.doCallback();
                    })
                );
                entity.attach(yesButton);
                entity.attach(noButton);
            },
            setupYesButton = function () {
                var spriteYesButton = settings.spriteYes ||
                    new Sprite({
                        image: Bento.assets.getImage(imageYes),
                        frameCountX: 1,
                        frameCountY: 3,
                        animations: {
                            up: {
                                frames: [0]
                            },
                            down: {
                                frames: [1]
                            },
                            inactive: {
                                frames: [2]
                            }
                        }
                    });
                var yesButton = new ClickButton({
                    name: 'buttonYes',
                    sfx: 'sfx_button',
                    sprite: spriteYesButton,
                    position: new Vector2(-5, buttonY),
                    updateWhenPaused: pauseLevel,
                    onClick: function () {
                        if (isMoving) {
                            return;
                        }
                        if (settings.onYes) {
                            Utils.Analytics.event('Dialog', 'dialog-' + settings.name + '-yes');
                            Utils.Firebase.logEvent('dialog_close', {
                                dialog_name: settings.name,
                                close_type: 'yes'
                            });
                            settings.onYes();
                        }
                        if (strings.length) {
                            // show next text
                            text.setText(strings.shift());
                            page += 1;
                            if (settings.onNextPage) {
                                settings.onNextPage(page);
                            }
                        } else {
                            moveOut();
                        }
                    }
                }).attach(
                    // back button pressed
                    new Utils.BackButton(function (data) {
                        data.entity.doCallback();
                    })
                );
                yesButton.name = 'buttonYes';
                entity.attach(yesButton);
            },
            moveIn = function () {
                new Tween({
                    from: viewport.x + viewport.width * 3 / 2,
                    to: viewport.x + viewport.width / 2 + 5,
                    'in': 30,
                    ease: 'exponential',
                    alpha: -6,
                    updateWhenPaused: pauseLevel,
                    'do': function (v, t) {
                        entity.position.x = v;
                    },
                    onComplete: function () {
                        isMoving = false;
                    }
                });
            },
            moveOut = function () {
                isMoving = true;
                new Tween({
                    from: entity.position.x,
                    to: viewport.x + viewport.width * 3 / 2 + 10,
                    'in': 30,
                    ease: 'exponential',
                    alpha: -6,
                    updateWhenPaused: pauseLevel,
                    'do': function (v, t) {
                        entity.position.x = v;
                    },
                    onComplete: function () {
                        isMoving = false;
                        if (settings.onClose) {
                            settings.onClose();
                        }
                        Bento.objects.remove(entity);
                    }
                });
            };

        if (!Utils.isArray(strings)) {
            strings = [strings];
        }

        text.setText(strings.shift());

        entity.attach(text);
        Bento.objects.attach(entity);

        // setup buttons
        if (type === 'yesno') {
            setupYesNoButtons();
        } else if (type === 'yes') {
            setupYesButton();
        }

        if (canClose) {
            entity.attach(buttonClose);
        }

        // move in
        moveIn();

        if (!settings.name) {
            console.warn("ERROR: Dialog must have a name");
        }

        return entity;
    };
});