/**
 * Module description
 */
bento.define('gui/team/page', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/drawinview'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    DrawInView
) {
    'use strict';
    return function (settings) {
        // pages are currently set up as 3 in a row
        // to a max of 6
        var page = new Entity({
            name: 'page',
            components: []
        });
        var items = [];
        var maxItems = settings.maxItems || 6;
        var offset = settings.offset || new Vector2(-45, -20);
        var spacing = settings.spacing || new Vector2(0, 2);
        var align = function () {
            var i = 0;
            var item;
            for (i = 0; i < items.length; ++i) {
                item = items[i];
                if (!item) {
                    continue;
                }
                item.position.x = offset.x + (i % 3) * (item.dimension.width + spacing.x);
                item.position.y = offset.y + Math.floor(i / 3) * (item.dimension.height + spacing.y);
            }
        };
        var activated = true;
        var alignTween = function () {
            var doTween = function (item, i) {
                new Tween({
                    from: item.position.x,
                    to: offset.x + (i % 3) * (item.dimension.width + spacing.x),
                    'in': 20,
                    ease: 'easeOutExpo',
                    updateWhenPaused: Utils.PauseLevels.TEAMEDIT,
                    'do': function (v, t) {
                        item.position.x = v;
                    }
                });
                new Tween({
                    from: item.position.y,
                    to: offset.y + Math.floor(i / 3) * (item.dimension.height + spacing.y),
                    'in': 20,
                    ease: 'easeOutExpo',
                    updateWhenPaused: Utils.PauseLevels.TEAMEDIT,
                    'do': function (v, t) {
                        item.position.y = v;
                    }
                });
            };
            Utils.forEach(items, function (item, i, l, breakLoop) {
                if (!item) {
                    breakLoop();
                }
                doTween(item, i);
            });
        };

        page.extend({
            iterate: function (callback) {
                var i = 0;
                for (i = 0; i < items.length; ++i) {
                    callback(items[i], i);
                }
            },
            addItem: function (item, prepend) {
                if (item.parent) {
                    console.log('warning: item already attached');
                    return;
                }
                if (prepend) {
                    items.unshift(item);
                } else {
                    items.push(item);
                }
                page.attach(item);
                alignTween();
            },
            getItems: function () {
                return items;
            },
            getLastItem: function () {
                return items[items.length - 1];
            },
            isFull: function () {
                return items.length >= maxItems;
            },
            removeItem: function (item) {
                var index = items.indexOf(item);
                if (index !== -1) {
                    items.splice(index, 1);
                    // items[index] = null;
                }

                page.remove(item);
                // ? we could realign the items
                alignTween();
            },
            swapItems: function (oldItem, newItem, oldPosition) {
                var index = items.indexOf(oldItem);
                var oldItemPosition = oldItem.getWorldPosition();
                var diffPosition = oldPosition.subtract(oldItemPosition);
                if (index !== -1) {
                    page.remove(oldItem);
                    items[index] = newItem;
                } else {
                    items.push(newItem);
                }
                page.attach(newItem);
                align();

                // tween position?
                new Tween({
                    from: newItem.position.x + diffPosition.x,
                    to: newItem.position.x,
                    'in': 10,
                    ease: 'easeOutExpo',
                    updateWhenPaused: Utils.PauseLevels.TEAMEDIT,
                    'do': function (v, t) {
                        newItem.position.x = v;
                    }
                });
                new Tween({
                    from: newItem.position.y + diffPosition.y,
                    to: newItem.position.y,
                    'in': 10,
                    ease: 'easeOutExpo',
                    updateWhenPaused: Utils.PauseLevels.TEAMEDIT,
                    'do': function (v, t) {
                        newItem.position.y = v;
                    }
                });
            },
            // activate and deactivate is for draggables performance
            activate: function () {
                var i, item;
                if (activated) {
                    return;
                }
                activated = true;
                // activate children
                for (i = 0; i < items.length; ++i) {
                    item = items[i];
                    if (!item) {
                        continue;
                    }
                    item.getComponent('draggable', function (draggable) {
                        draggable.activate();
                    });
                    // page.attach(item);
                }
            },
            deactivate: function () {
                var i, item;
                if (!activated) {
                    return;
                }
                activated = false;
                // deactivate children
                for (i = 0; i < items.length; ++i) {
                    item = items[i];
                    if (!item) {
                        continue;
                    }
                    item.getComponent('draggable', function (draggable) {
                        draggable.deactivate();
                    });
                    // page.remove(item);

                }
            }
        });

        page.attach(new DrawInView({}));

        return page;
    };
});