/**
 * Initialization module
 */
bento.define('init', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/animation',
    'bento/components/sprite',
    'bento/eventsystem',
    'bento/tween',
    'autoresize',
    'utils',
    'dummy',
    'modules/betatest',
    'devsaver',
    'modules/fbnotification',
    'modules/tenjin',
    'modules/ironsource'
], function (
    Bento,
    Vector2,
    Rectangle,
    Animation,
    Sprite,
    EventSystem,
    Tween,
    AutoResize,
    Utils,
    Dummy,
    BetaTest,
    DevSaver,
    FBNotification,
    Tenjin,
    IronSource
) {
    'use strict';
    return function () {
        /**
         * Clears screen with black every tick (android only)
         */
        var canvasDimension = Bento.getViewport();
        // var clear = function (data) {
        //     data.renderer.begin();
        //     data.renderer.fillRect([0, 0, 0, 1], 0, 0, canvasDimension.width, canvasDimension.height);
        //     data.renderer.flush();
        // };
        // if (Utils.isNativeAndroid()) {
        //     EventSystem.on('preDraw', clear);
        // }

        // cocoonjs
        if (Utils.isCocoonJS() && window.Cocoon) {
            window.Cocoon.Utils.setAntialias(false);
        }
        /**
         * Init singletons
         */
        Utils.Firebase.init();
        Utils.Analytics.init();
        if (Utils.isCocoonJS()) {
            if (Utils.isNativeIos()) {
                Utils.GameCenter.login();
            } else if (Utils.isNativeAndroid()) {
                Utils.GooglePlay.init(function () {
                    // its actually safer to not login during the init???
                });
                // after logging in immediately check player stats
                Utils.GooglePlay.login(Utils.checkIfHighSpender);
            }
            // if user has allowed notifications, initialize notification system
            if (Bento.saveState.load('notificationsAllowed', false)) {
                Utils.Notification.init();
            }
        }
        // analytics
        (function () {
            var bootup = Bento.saveState.load('bootup', 0);
            var sessions = Bento.saveState.load('sessions', 0);
            bootup += 1;
            sessions += 1;

            Bento.saveState.save('bootup', bootup);
            Bento.saveState.save('sessions', sessions);
            Bento.saveState.save('sessionTime', Date.now());
            if (window.device) {
                Utils.Analytics.event('UI', 'model', undefined, window.device.model.model);
                Utils.Analytics.event('UI', 'operating_system', undefined, window.device.model.os);
            } else {
                Utils.Analytics.event('UI', 'model', undefined, 'unknown');
            }
            Utils.Analytics.event('UI', 'totalBootup', bootup);

            // this game does not require a pause to be triggered after suspending
            document.addEventListener('pause', function () {
                var now = Date.now();
                var diff = now - Bento.saveState.load('sessionTime', Date.now());
                Utils.Analytics.event('Game', 'sessionTime_min', Math.floor(diff / 1000 / 60));
                Bento.saveState.save('sessionTime', now);
            });
            document.addEventListener('resume', function () {
                sessions = Bento.saveState.load('sessions', 0);
                sessions += 1;
                Bento.saveState.save('sessions', 0);

                // reset iOS badge number when opening app
                Utils.Notification.resetBadgeNumber();

                // Standard Firebase event: By logging this event when an App is moved to the foreground, developers can understand how often users leave and return during the course of a Session
                Utils.Firebase.logEvent(Utils.Firebase.Events.APP_OPEN);
            });
        })();
        /**
         * Init game
         */
        // unlock new planets for people who have beat the game before the first content update
        if (
            Utils.hasDefeatedBoss('gingyking') &&
            (
                Bento.saveState.load('unlockedPlanetNames', ['starti']).indexOf('pirate') < 0 ||
                Bento.saveState.load('unlockedPlanetNames', ['starti']).indexOf('lavalli') < 0
            )
        ) {
            var planetsToUnlock = Bento.saveState.load('planetsToUnlock', []);
            // 'ending1' should not be in planetsToUnlock as it's not a planet and the game won't unlock any other planets
            if (planetsToUnlock.indexOf('ending1') > -1) {
                planetsToUnlock.splice(planetsToUnlock.indexOf('ending1'), 1);
            }
            // queue up the two new planets for unlocking
            if (Bento.saveState.load('unlockedPlanetNames', ['starti']).indexOf('pirate') < 0) {
                planetsToUnlock.push('pirate');
            }
            if (Bento.saveState.load('unlockedPlanetNames', ['starti']).indexOf('lavalli') < 0) {
                planetsToUnlock.push('lavalli');
            }
            Bento.saveState.save('planetsToUnlock', planetsToUnlock);

            console.log('Player has beat candy planet before upgrade. Planets to unlock:');
            for (var i = 0; i < planetsToUnlock.length; ++i) {
                console.log(planetsToUnlock[i]);
            }
        }
        /**
         * Haptic feedback: every button should give feedback
         */
        var startHapticFeedback = function () {
            if (!Utils.TapticEngine) {
                console.log('WARNING: TapticEngine does not exist in Utils');
                return;
            }
            EventSystem.on('clickButton-onButtonDown', function (data) {
                Utils.TapticEngine.impact({
                    style: 'light'
                });
            });
            EventSystem.on('toggleButton-onToggle', function (data) {
                Utils.TapticEngine.selection();
            });
            // deprecated events
            EventSystem.on('clickButton', function (data) {
                Utils.TapticEngine.impact({
                    style: 'light'
                });
            });
            // for testing purposes: replace the official API with unofficial for iphone6
            if (Bento.isDev() && window.device && window.device.model === 'iPhone8,1') {
                Utils.TapticEngine.initUnofficial();
            }
            // Utils.toggleTapticEngine(Bento.saveState.load(Utils.SaveKeys.TAPTICENGINE, true));
        };

        startHapticFeedback();
        
        // init tenjin
        Tenjin.init({
            apiKey: 'TVITWX4WGSZGBSL58QGWY7XLBYXIAK4B',
            optIn: true, // will we ever opt out?
            onComplete: function (error) {
                if (!error) {
                    console.log('Tenjin initialized');
                } else {
                    console.warn('WARNING: ' + error);
                }
            }
        });

        // init ironsource
        console.log('Ironsource initialize');
        IronSource.initialize({
            appKey: Utils.isNativeIos() ? 'a352d605' : 'a353829d',
            clientCallback: false,
            autoLoadInterstitials: true
        }, function () {
            IronSource.setUserConsent(true);
            console.log('Ironsource validate integration');
            window.plugins.ironsource.validateIntegration(function (data) {
                console.log(data);
            }, function (data) {
                console.log(data);
            });
        }, function (e) {});

        Bento.audio.muteSound(Bento.saveState.load('muteSound', false));
        Bento.audio.muteMusic(Bento.saveState.load('muteMusic', false));
        Bento.assets.load('preloader', function (err) {
            // v1.3.1-1 patch: lionfu
            /*var assets = Bento.assets.getAssets();
            assets.json['names/de'].strings.lionfu = assets.json['names/de'].strings.loinfu || assets.json['names/de'].strings.lionfu;
            assets.json['names/en'].strings.lionfu = 'Lionfu';
            assets.json['names/es'].strings.lionfu = assets.json['names/es'].strings.loinfu || assets.json['names/es'].strings.loinfu;
            assets.json['names/fr'].strings.lionfu = assets.json['names/fr'].strings.loinfu || assets.json['names/fr'].strings.loinfu;
            assets.json['names/it'].strings.lionfu = assets.json['names/it'].strings.loinfu || assets.json['names/it'].strings.loinfu;
            assets.json['names/ja'].strings.lionfu = assets.json['names/ja'].strings.loinfu || assets.json['names/ja'].strings.loinfu;
            assets.json['names/ko'].strings.lionfu = assets.json['names/ko'].strings.loinfu || assets.json['names/ko'].strings.loinfu;
            assets.json['names/pt'].strings.lionfu = assets.json['names/pt'].strings.loinfu || assets.json['names/pt'].strings.loinfu;
            assets.json['names/ru'].strings.lionfu = assets.json['names/ru'].strings.loinfu || assets.json['names/ru'].strings.loinfu;
            assets.json['names/zh'].strings.lionfu = assets.json['names/zh'].strings.loinfu || assets.json['names/zh'].strings.loinfu;
            assets.json['names/zt'].strings.lionfu = assets.json['names/zt'].strings.loinfu || assets.json['names/zt'].strings.loinfu;*/

            // after langauage jsons are loaded, initialize
            Utils.Localization.initLanguage();
            console.log('show preloader!')
            Bento.screens.show('screens/preloader');
        });

        Bento.saveState.save('currentVersion', window.VERSION);


        // test: schedule notification 24 hours ahead
        // FBNotification.schedule(24, {
        //     message: 'Test',
        //     title: 'Test',
        //     buttonText: 'Play',
        //     imageUrl: 'https://s3.amazonaws.com/lucky-kat/combocritters/messenger.jpg'
        // }, function (error, messageDataStr) {
        //     if (error) {
        //         console.error(messageDataStr);
        //         return;
        //     }
        //     // save messageData which can be used to delete the notification
        //     Bento.saveState.save('notification24', messageDataStr);
        //     console.log(messageDataStr);
        // });

        // FBNotification.remove(Bento.saveState.load('notification24', '{}'), function (error, msg) {
        //     if (error) {
        //         console.error(msg);
        //         return;
        //     }
        //     console.log('cancelled');
        // });

        // allow dev-saving on desktop
        if (Bento.isDev() && !window.Cocoon) {
            DevSaver();
        }
    };
});