/**
 * Arealoader takes a Tiled json and calls subTiled to load specific parts of the map on screen
 */
bento.define('modules/arealoader', [
    'bento',
    'bento/components/sprite',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'subtiled',
    'utils',
    'modules/spatialhashgrid'
], function (
    Bento,
    Sprite,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    SubTiled,
    Utils,
    SpatialHashGrid
) {
    'use strict';
    var id = 0;

    return function (settings) {
        /* settings = {
            json: Json asset of the area to load
            offset: Vector2
            areaLoadTrigger: Entity // reference to the areaLoadTrigger
        }
        */
        var i,
            currentSection,
            sections = [],
            sectionsToLoad = [],
            /*{
                tiled: Tiled,
                dimension: Rectangle,
                loadRect: Rectangle, // size of tiled rectangle to load (in tiles, not pixels)
                loadTriggerSize: Number // factor size of load trigger
                unloadTriggerSize: Number // factor size of load trigger
                tiledRect: 
                loaded: Boolean,
                cachedLayers: Boolean
            },*/
            json = Bento.assets.getJson(settings.json),
            player,
            areaLoadTrigger = settings.areaLoadTrigger || null,
            tileWidth = json.tilewidth,
            tileHeight = json.tileheight,
            loadRectangle = settings.loadRect || new Rectangle(0, 0, 16, 16),
            loadTriggerSize = settings.loadTriggerSize || 1.5,
            unloadTriggerSize = settings.unloadTriggerSize || 2,
            fullLevelSize,
            width,
            height,
            loadedParalax = false,
            loadedSections = 0,
            totalSections = 0,
            didCallback = false,
            loadGrid,
            loadedSectionsArray = [],
            entity = new Entity({
                name: settings.json,
                position: settings.offset || new Vector2(0, 0),
                updateWhenPaused: Utils.PauseLevels.ALWAYS,
                family: ['areaLoader']
            }).extend({
                /**
                 * Returns a rectangle with the total size of the level.
                 * @function
                 * @instance
                 * @name getFullLevelSize
                 */
                getFullLevelSize: function () {
                    // cache size
                    if (!fullLevelSize) {
                        fullLevelSize = new Rectangle(
                            entity.position.x,
                            entity.position.y,
                            json.width * tileWidth,
                            json.height * tileWidth
                        );
                    }

                    return fullLevelSize;
                },
                /**
                 * Returns true if player is inside the vicinity.
                 */
                isPlayerHere: function () {
                    var rect;
                    if (!areaLoadTrigger) {
                        return;
                    }
                    rect = this.getFullLevelSize().clone().grow(-16);

                    return areaLoadTrigger.getBoundingBox().intersect(rect);
                }
            }),
            /**
             * Checks if areaLoadTrigger is intersecting with any of the level load triggers. If so, loads the corresponding section.
             * @function
             * @instance
             * @name sectionLoader
             */
            sectionLoader = function () {
                var tiled,
                    i,
                    viewport = Bento.getViewport(),
                    collidesWithLoad = false;

                if (!player) {
                    // player must be loaded
                    Bento.objects.get('player', function (obj) {
                        player = obj;
                    });
                    return;
                }
                if (!areaLoadTrigger) {
                    // make one
                    areaLoadTrigger = new Entity({
                        z: 0,
                        name: 'areaLoadTrigger',
                        position: new Vector2(0, 0),
                        originRelative: new Vector2(0, 0),
                        components: [{
                            update: function () {
                                areaLoadTrigger.position.x = viewport.x + viewport.width / 2;
                                areaLoadTrigger.position.y = viewport.y + viewport.height / 2;
                            }
                        }]
                    });
                    areaLoadTrigger.dimension.width = 1;
                    areaLoadTrigger.dimension.height = 1;
                    areaLoadTrigger.position.x = viewport.x + viewport.width / 2;
                    areaLoadTrigger.position.y = viewport.y + viewport.height / 2;
                    Bento.objects.attach(areaLoadTrigger);
                }

                // load
                // loadGrid.collidesWith(areaLoadTrigger, new Vector2(0, 0), function (other) {
                //     var section = other.section;
                //     if (!section.loaded) {
                //         // console.log('load section', section.id)
                //         totalSections += 1;
                //         section.loaded = true;
                //         sectionsToLoad.push(section);
                //         loadedSectionsArray.push(section);
                //     }
                // }, true);
                var fullLevelSize = entity.getFullLevelSize();
                var sectionCountX = fullLevelSize.width / (loadRectangle.width * tileWidth);
                var sectionCountY = fullLevelSize.height / (loadRectangle.height * tileHeight);
                var sectionX = Math.floor(areaLoadTrigger.position.x / (loadRectangle.width * tileWidth));
                var sectionY = Math.floor(areaLoadTrigger.position.y / (loadRectangle.height * tileHeight));
                var x, y, loopX, loopY;
                var tileRect;
                var section;
                var split;
                var minX = sectionX - 1;
                var maxX = sectionX + 1;
                var minY = sectionY - 3;
                var maxY = sectionY + 1;
                // load the surrounding sections
                for (x = minX; x < maxX + 1; ++x) {
                    for (y = minY; y < maxY + 1; ++y) {
                        // not loaded yet
                        if (!areasLoaded[x + ',' + y]) {
                            loopX = (x % sectionCountX);
                            loopY = (y % sectionCountY);
                            while (loopX < 0) {
                                loopX += sectionCountX;
                            }
                            while (loopY < 0) {
                                loopY += sectionCountX;
                            }
                            tileRect = new Rectangle(
                                loopX * loadRectangle.width,
                                loopY * loadRectangle.height,
                                loadRectangle.width,
                                loadRectangle.height
                            );
                            section = {
                                tiled: null,
                                tileRect: tileRect,
                                dimension: new Rectangle(
                                    tileRect.x * tileWidth,
                                    tileRect.y * tileHeight,
                                    tileRect.width * tileWidth,
                                    tileRect.height * tileHeight
                                ),
                                loaded: true,
                                offset: new Vector2(
                                    Math.floor(x / sectionCountX) * fullLevelSize.width,
                                    Math.floor(y / sectionCountY) * fullLevelSize.height
                                )
                            };
                            totalSections += 1;
                            loadTiled(section);
                            areasLoaded[x + ',' + y] = section;
                        }
                    }
                }
                // unload outer sections
                for (section in areasLoaded) {
                    if (!areasLoaded.hasOwnProperty(section)) {
                        continue;
                    }
                    split = section.split(',');
                    x = parseInt(split[0]);
                    y = parseInt(split[1]);
                    if (x < minX || x > maxX || y < minY || y > maxY) {
                        // console.log('unload', x, y);
                        if (areasLoaded[section].tiled) {
                            if (settings.onSectionUnloaded) {
                                settings.onSectionUnloaded(areasLoaded[section].tiled);
                            }
                            areasLoaded[section].tiled.remove();
                        }
                        delete areasLoaded[section];
                    }
                }
            },
            areasLoaded = {},
            unloadSection = function (section, index) {
                if (!section.loaded) {
                    return;
                }
                if (!areaLoadTrigger.getBoundingBox().intersect(section.unloadRect)) {
                    if (settings.onSectionUnloaded) {
                        settings.onSectionUnloaded(section.tiled);
                    }
                    if (section.tiled) {
                        section.tiled.remove();
                        section.tiled = null;
                    }
                    section.loaded = false;
                    // console.log('unload', section.id);
                    // remove from loaded array
                    loadedSectionsArray[index] = null;
                }
            },
            loadTiled = function (section) {
                var tiled;

                // remove could be called before tiled actually gets loaded
                if (!section.loaded) {
                    console.log('WARNING: tiled loading skipped');
                    return;
                }

                tiled = new SubTiled({
                    json: json,
                    name: settings.json, // also pass json name, passed on to spawned objects
                    spawn: true,
                    rectangle: section.tileRect,
                    offset: section.offset || new Vector2(0, 0), //entity.position,
                    cachedLayers: settings.cachedLayers,
                    onLoad: function (tiledSettings, objects) {
                        loadedSections += 1;
                        if (totalSections > 1 && loadedSections === totalSections && settings.onLoad && !didCallback) {
                            didCallback = true;
                            settings.onLoad(totalSections);

                        }

                    }
                });
                section.tiled = tiled;
                if (settings.onSectionLoaded) {
                    settings.onSectionLoaded(section.tileRect, tiled, totalSections);
                }
            },
            /**
             * Adds the name of the section, rectangles that trigger loading and unloading, and a boolean that keeps track of if the sections is loaded.
             * @function
             * @instance
             * @name addToLoadTriggers
             */
            addToLoadTriggers = function (rectangle) {
                // get x and y order of section from json name
                var rect = new Rectangle(
                        rectangle.x * tileWidth,
                        rectangle.y * tileHeight,
                        rectangle.width * tileWidth,
                        rectangle.height * tileHeight
                    ),
                    loadRect = rect.clone().offset(entity.position),
                    unloadRect = rect.clone().offset(entity.position),
                    loadTrigger = new Entity({}),
                    unloadTrigger = new Entity({}),
                    section;

                // the rectangles that trigger loading are 1.3x as big as the section
                // the rectangles that trigger unloading are 1.5x as big and are later check for non-collision
                loadRect.grow(loadTriggerSize * rect.width, false, true);
                loadRect.grow(loadTriggerSize * rect.height, true, false);
                unloadRect.grow(unloadTriggerSize * rect.width, false, true);
                unloadRect.grow(unloadTriggerSize * rect.height, true, false);

                section = {
                    id: id++,
                    tiled: null,
                    tileRect: rectangle,
                    dimension: rect,
                    loadRect: loadRect,
                    unloadRect: unloadRect,
                    loaded: false
                };
                sections.push(section);

                // init spatial hashgrid if needed
                if (!loadGrid) {
                    loadGrid = new SpatialHashGrid({
                        cellSize: new Vector2(rect.width, rect.height)
                    });
                }

                // set position and dimension of (un)load trigger
                loadTrigger.position.x = loadRect.x;
                loadTrigger.position.y = loadRect.y;
                loadTrigger.dimension.width = loadRect.width;
                loadTrigger.dimension.height = loadRect.height;
                // remember a reference to the section
                loadTrigger.extend({
                    section: section
                });
                loadGrid.insert(loadTrigger);
            };
        entity.attach({
            update: function (data) {
                var i;
                sectionLoader();

                // if (sectionsToLoad.length > 0 && entity.timer % 5 === 0) {
                //     loadTiled(sectionsToLoad.shift());
                // }

                // // unload if too far away
                // for (i = 0; i < loadedSectionsArray.length; ++i) {
                //     unloadSection(loadedSectionsArray[i], i);
                // }
                // // clean up
                // for (i = loadedSectionsArray.length - 1; i >= 0; --i) {
                //     if (loadedSectionsArray[i] === null) {
                //         loadedSectionsArray.splice(i, 1);
                //     }
                // }
            }
        });

        // preload entire map
        if (settings.cachedLayers) {
            width = Math.ceil(json.width / loadRectangle.width);
            height = Math.ceil(json.height / loadRectangle.height);
            for (i = 0; i < width * height; ++i) {
                new SubTiled({
                    json: json,
                    name: settings.json, // also pass json name, passed on to spawned objects
                    rectangle: new Rectangle(
                        (i % width) * loadRectangle.width,
                        Math.floor(i / width) * loadRectangle.height,
                        loadRectangle.width,
                        loadRectangle.height
                    ),
                    cachedLayers: settings.cachedLayers,
                    spawn: false, // don't spawn anything, just for preloading
                    onLoad: function () {}
                });
            }
        }

        return entity;
    };
});