/**
 * Cocoon save file migration
 * - Requires cordova-plugin-file
 * - Optional but recommended is cordova-plugin-nativestorage
 * @moduleName CocoonMigration
 */
bento.define('modules/cocoonmigration', [
    'bento',
    'bento/utils',
    'modules/cocoonmigrationlib'
], function (
    Bento,
    Utils,
    Cocoonmigrationlib
) {
    'use strict';
    /**
     * Locates and reads ls.cdf
     */
    var readLocalStorage = function (onComplete) {
        if (!window.cordova) {
            onComplete(null);
            return;
        }
        var dataDirectory = window.cordova.file.applicationStorageDirectory;
        var readFromFile = function (fileName, onSuccess, onFail) {
            var completePath = dataDirectory + 'Library/Application Support/' + fileName;
            if (Utils.isNativeAndroid()) {
                // iOS and Android file system layouts are different
                completePath = dataDirectory + 'files/' + fileName;
            }

            window.resolveLocalFileSystemURL(completePath, function (fileEntry) {
                fileEntry.file(function (file) {
                    var reader = new window.FileReader();
                    reader.onloadend = function (e) {
                        var result = this.result;
                        if (onSuccess) {
                            onSuccess(result);
                        }
                        return result;
                    };
                    reader.readAsDataURL(file);
                }, onFail);

            }, onFail);
        };

        readFromFile('ls.cdf', function (data) {
            onComplete(null, data);
        }, function () {
            onComplete('ERROR: could not read Cocoon localStorage');
        });
    };
    /**
     * Rename ls.cdf to _ls.cdf
     */
    var renameLocalStorage = function (onComplete, revert) {
        var dataDirectory = window.cordova.file.applicationStorageDirectory;
        var rename = function (fileName, newName, onSuccess, onFail) {
            var directory = dataDirectory + 'Library/Application Support/';
            var completePath = directory + fileName;
            if (Utils.isNativeAndroid()) {
                // iOS and Android file system layouts are different
                directory = dataDirectory + 'files/';
                completePath = directory + fileName;
            }

            window.resolveLocalFileSystemURL(completePath, function (fileEntry) {
                window.resolveLocalFileSystemURL(directory, function (directoryEntry) {
                    fileEntry.moveTo(directoryEntry, newName, onSuccess, onFail);
                }, onFail);
            }, onFail);
        };
        var name = 'ls.cdf';
        var target = '_ls.cdf';

        if (revert) {
            name = [target, target = name][0];
        }

        console.log('Renaming localStorage');
        rename(name, target, function (data) {
            console.log('Successfully renamed file.');
            onComplete(null);
        }, function () {
            onComplete('ERROR: did not rename anything');
        });
    };
    /**
     * Used this function to read the file storage and locate the localstorage file, not much use now
     */
    var search = function () {
        var dataDirectory = window.cordova.file.applicationStorageDirectory;
        var enumerateFiles = function (directoryEntry) {
            // var dirReader = file.root.createReader();
            var dirReader = directoryEntry.createReader();
            var entries = [];
            var toArray = function (list) {
                return Array.prototype.slice.call(list || [], 0);
            };
            var listResults = function (entr) {
                entr.forEach(function (entry, i) {
                    console.log((entry.isDirectory ? '/' : '') + entry.name);
                });
            };
            // Call the reader.readEntries() until no more results are returned.
            var readEntries = function () {
                dirReader.readEntries(function (results) {
                    if (!results.length) {
                        listResults(entries.sort());
                    } else {
                        entries = entries.concat(toArray(results));
                        readEntries();
                    }
                }, function () {
                    console.log('ERROR');
                });
            };

            readEntries(); // Start reading dirs.
        };
        window.resolveLocalFileSystemURL(dataDirectory + '', function (dirEntry) {
            console.log('enumerating in ' + dataDirectory);
            enumerateFiles(dirEntry);
        }, function (f, error) {
            console.log('resolve error ' + dataDirectory, JSON.stringify(f), error);
        });
    };
    /**
     * Main module
     */
    var CocoonMigration = {
        /**
         * @snippet CocoonMigration.migrateToNativeStorage()|Snippet
CocoonMigration.migrateToNativeStorage({
    doOnce: true, // will rename ls.cdf afterwards
    onKey: function (key, valueAsString) {
        // note: key will include the id if present, for example "beatstreet/coins"
        // return true to ignore key for migration
        return false;
    },
    onComplete: function (error, keysMigrated) {
        // completion is asynchronous!
        // note: make sure migrate from Cocoon first BEFORE you call NativeStorage.init() (--> so call NativeStorage.init here)
    }
});
         */
        migrateToNativeStorage: function (settings) {
            // first read ls.cdf
            var onComplete = settings.onComplete;
            var doOnce = settings.doOnce;
            var onKey = settings.onKey;
            var parsePList = function (plistText) {
                var parser = new window.DOMParser();
                var xmlDoc = parser.parseFromString(plistText, "text/xml");
                var dict = xmlDoc.getElementsByTagName("dict");

                var currentKey = '';
                var currentString = '';
                var keysToMigrate = 0;
                var migrated = 0;
                var checkDone = function () {
                    if (migrated === keysToMigrate) {
                        if (doOnce) {
                            renameLocalStorage(function (error) {
                                if (error) {
                                    console.log(error);
                                }
                                onComplete(null, migrated);
                            });
                        } else {
                            onComplete(null, migrated);
                        }
                    }
                };
                if (dict && dict.length) {
                    Utils.forEach(dict[0].childNodes, function (childNode) {
                        var shouldIgnore = false;
                        if (childNode.nodeName === 'key') {
                            // first collect the key
                            currentKey = childNode.textContent;
                        } else if (childNode.nodeName === 'string') {
                            // then get the value
                            currentString = childNode.textContent;

                            if (onKey) {
                                shouldIgnore = onKey(currentKey, currentString);
                            }

                            // once we have both, migrate it to nativeStorage
                            if (window.NativeStorage && !shouldIgnore) {
                                keysToMigrate += 1;
                                window.NativeStorage.setItem(currentKey, currentString, function () {
                                    migrated += 1;
                                    checkDone();
                                }, function (error) {
                                    console.log('Migration error for key ' + currentKey + ':' + error);
                                    // increment anyway
                                    migrated += 1;
                                    checkDone();
                                });
                            }
                        }
                    });
                } else {
                    // nothing to do
                    checkDone();
                }
            };

            // read the localStorage file
            readLocalStorage(function (error, encryptedData) {
                if (error || !encryptedData) {
                    onComplete(error, 0);
                    return;
                }
                try {
                    // will have to be decrypted
                    var decrypted = window.DecryptData(
                        encryptedData.replace('data:*/*;base64,', '').replace('data:application/x-cdf;base64,', ''),
                        "gdfsTT464fga5."
                    );
                    // parse content as XML/PList
                    parsePList(decrypted);
                } catch (e) {
                    onComplete(e, 0);
                }
            });
        },
        search: function () {
            // for debugging only
            search();
        },
        revert: function () {
            // for debugging only
            renameLocalStorage(function (error) {}, true);
        }
    };
    return CocoonMigration;
});