/**
 * Module for using IronSource Ads in Bento.
 TODO: Banners
 TODO: Offer Walls
 TODO: Other Callbacks(?)
 * @moduleName IronSource
 */
bento.define('modules/ironsource', [
    'bento',
    'bento/utils',
    'bento/eventsystem'
], function (
    Bento,
    Utils,
    EventSystem
) {
    // ------ FUNCTIONS ------
    var nothing = function () {
        // do nothing
    };
    var error = function (err) {
        if (!err) {
            return;
        }
        console.log('Iron Source Error: ', err);
    };
    var prepareLoop = function () {
        // keeps trying to prepare until it succeeds
        IronSourceModule.prepareInterstitial(nothing, prepareLoop);
    };


    // ------ VARS ------
    // Interstitial Variables
    var autoLoadInterstitials = true;
    var isProcessingInterstitial = false;
    var interstitialLoadCallback;
    var interstitialLoadFailedCallback;
    var interstitialCompleteCallback;

    // Rewarded Variables
    var isProcessingRewarded = false;
    var doGiveReward = false;
    var rewardedCallback;


    // ------ MODULE ------
    // Fallback Callbacks Module
    var IronSourceModule = {
        initialize: nothing,
        setUserConsent: nothing,
        prepareInterstitial: nothing,
        canShowInterstitial: function (a, b) {
            b();
        },
        showInterstitial: nothing,
        canShowRewarded: function (a, b) {
            b();
        },
        showRewarded: nothing
    };

    // Check if we can access iron source, if we can, get it ready
    if (window.plugins && window.plugins.ironsource) {
        var IronSource = window.plugins.ironsource;
        IronSourceModule = {
            /**
            * Initializes the SDK and creates the events

            * @param {Object} settings -
                options = {
                    appKey: 'appKey' string from IronSource console
                    userId: 'userId' (optional) string for the userId
                    clientCallback: boolean for whether to use client callbacks for offerwalls, useless right now
                    autoLoadInterstitials: boolean for whether we want to automatically load interstitials in the same way rewarded ads load
                }

            * @snippet IronSource.initialize
            IronSource.initialize({
                appKey: 'appKey',
                userId: 'userId',
                clientCallback: false,
                autoLoadInterstitials: true
            }, function () {
                //success
            }, function (e) {
                //failure
            });
            */
            initialize: function (options, success, failure) {
                if (Utils.isDefined(options.autoLoadInterstitials)) {
                    autoLoadInterstitials = options.autoLoadInterstitials;
                }
                if (!options.appKey) {
                    console.warn('Iron Source Initialization failed!', 'No App Key!');
                }
                IronSource.initSdk({
                    appKey: options.appKey || '',
                    userId: options.userId || '',
                    clientCallback: Utils.isDefined(options.clientCallback) ? options.clientCallback : false,
                }, function () {
                    console.log('Iron Source Initialized!');
                    if (success) {
                        success();
                    }
                    if (autoLoadInterstitials) {
                        prepareLoop();
                    }
                }, function (e) {
                    console.warn('Iron Source Initialization failed!', e);
                    if (failure) {
                        failure(e);
                    }
                });

            },



            /**
            * This function is used to manually set the user consent for GDPR.

            * @snippet IronSource.setUserConsent
                IronSource.setUserConsent(true);
            */
            setUserConsent: function (userConsent) {
                IronSource.setUserConsent(userConsent, nothing, error);
            },



            /**
            * Manually prepares an Interstital to be shown, MUST be called before showInterstitial if autoLoadInterstitials is false

            * @param {Object} settings -
                success: callback (async),
                failure: callback (sync)

            * @snippet IronSource.prepareInterstitial
                IronSource.prepareInterstitial (function () {
                    //success
                }, function (e) {
                    //failure
                });
            */
            prepareInterstitial: function (success, failure) {
                interstitialLoadCallback = success;
                interstitialLoadFailedCallback = failure;
                IronSource.prepareInterstitialView(nothing, error);
            },



            /**
            * Checks whether we can show an interstitial

            * @param {Object} settings -
                canShowCallback: callback (async),
                cannotShowCallback: callback(reason) (async)

            * @snippet IronSource.canShowInterstitial
                IronSource.canShowInterstitial (function () {
                    // can show Interstitial
                }, function (e) {
                    // can not show Interstitial
                });
            */
            canShowInterstitial: function (canShowCallback, cannotShowCallback) {
                // annoyingly this is asynchoronous
                IronSource.isInterstitialReady(function (isInterstitialReady) {
                    if (isInterstitialReady) {
                        canShowCallback();
                    } else {
                        cannotShowCallback();
                    }
                }, function (e) {
                    if (cannotShowCallback) {
                        cannotShowCallback(e);
                    } else {
                        error(e);
                    }
                });
            },



            /**
            * Shows a prepared interstitial

            * @param {Object} settings -
                success: callback (sync),
                failure: callback(reason) (sync)

            * @snippet IronSource.showInterstitial
                IronSource.showInterstitial (function () {
                    // Interstitial completed
                }, function (e) {
                    // Interstitial failed to show
                });
            */
            showInterstitial: function (onComplete, failure, tag) {
                // callback that fires on a failure
                var onFailure = function (e) {
                    isProcessingInterstitial = false;
                    if (failure) {
                        failure(e);
                    } else {
                        error(e);
                    }
                };

                if (!isProcessingInterstitial) {
                    isProcessingInterstitial = true;
                    interstitialCompleteCallback = onComplete;
                    IronSourceModule.canShowInterstitial(function (didSucceed) {
                        IronSource.showInterstitialView(tag || '', nothing, onFailure);
                    }, onFailure);
                } else {
                    if (failure) {
                        failure('Cannot show ad at this time');
                    } else {
                        error('Cannot show ad at this time');
                    }
                }
            },



            /**
            * Checks whether we can show an rewarded

            * @param {Object} settings -
                success: callback (sync),
                failure: callback (sync)

            * @snippet IronSource.canShowRewarded
                IronSource.canShowRewarded (function () {
                    // can show rewarded
                }, function (e) {
                    // can not show rewarded
                });
            */
            canShowRewarded: function (canShowCallback, cannotShowCallback) {
                // annoyingly this is asynchoronous
                IronSource.hasRewardedVideo(function (hasRewardedVideo) {
                    if (hasRewardedVideo) {
                        canShowCallback();
                    } else {
                        cannotShowCallback();
                    }
                }, function (e) {
                    if (cannotShowCallback) {
                        cannotShowCallback(e);
                    } else {
                        error(e);
                    }
                });
            },



            /**
            * Shows a rewarded ad

            * @param {Object} settings -
                success: callback (sync),
                failure: callback (sync)

            * @snippet IronSource.showRewarded
                IronSource.showRewarded (function () {
                    // Rewarded shown and completed
                    //give reward
                }, function (e) {
                    // Interstitial failed to show
                    // rewarded failed to show, or the user cancelled watching
                });

            */
            showRewarded: function (onComplete, failure, tag) {
                // callback that fires on a failure
                var onFailure = function (e) {
                    isProcessingRewarded = false;
                    if (failure) {
                        failure(e);
                    } else {
                        error(e);
                    }
                };

                // if we're not already processing a rewarded ad, start a new process
                if (!isProcessingRewarded) {
                    //start process, set callback, reset bool
                    isProcessingRewarded = true;
                    rewardedCallback = onComplete;
                    doGiveReward = false;

                    //double check we can show a rewarded ad
                    IronSourceModule.canShowRewarded(function (didSucceed) {
                        IronSource.showRewardedVideo(tag || '', nothing, onFailure);
                    }, onFailure);
                } else {
                    // we're already processing an ad
                    if (failure) {
                        failure('Cannot show ad at this time');
                    } else {
                        error('Cannot show ad at this time');
                    }
                }
            }
        };

        // Good to go!
        console.log("IronSource Ready!");
    } else {

        //Iron Source doesn't exist, so we don't overwrite the module
        console.warn("WARNING: IronSource not found!");
    }

    // ------ LISTENERS ------
    //Interstitial ad close Listener
    document.addEventListener('interstitialDidClose', function (data) {
        // If we're waiting for a close
        if (isProcessingInterstitial) {
            isProcessingInterstitial = false;
            // fire success
            if (interstitialCompleteCallback) {
                interstitialCompleteCallback();
            }
            // prepare next interstitial
            if (autoLoadInterstitials) {
                prepareLoop();
            }
        }
    }, false);

    //Interstitial Load Ad Listener
    document.addEventListener('interstitialDidLoad', function (data) {
        if (interstitialLoadCallback) {
            interstitialLoadCallback(data);
        }
    }, false);

    //Interstitial Load Ad Failure Listener
    document.addEventListener('interstitialDidFailToLoadWithError', function (data) {
        if (interstitialLoadFailedCallback) {
            interstitialLoadFailedCallback(data);
        }
    }, false);

    // Rewarded ad close Listener
    document.addEventListener('rewardedVideoDidClose', function (data) {
        // if we're waiting for a close event
        if (isProcessingRewarded) {
            isProcessingRewarded = false;
            // if a reward function exists
            if (rewardedCallback) {
                // if we got the all clear for a reward, give it. Otherwise do nothing and wait for the reward listener
                if (doGiveReward) {
                    rewardedCallback(data);
                    doGiveReward = false;
                }
            }
        }
    }, false);

    // Reward Listener
    document.addEventListener('didReceiveRewardForPlacement', function (data) {
        if (isProcessingRewarded) {
            // if we're in the middle of processing a rewarded ad, allow the reward on close
            doGiveReward = true;
        } else {
            // if we get this callback and we're not processing an ad, the ad has already closed before we got this callback.
            // In this case, we just give them the reward as they have to have watched the ad to get it.
            rewardedCallback(data);
            doGiveReward = false;
        }
    }, false);

    return IronSourceModule;
});