/**
 * Spatial hash grid
 * <br>Exports: Function
 * @module bento/spatialhashgrid
 * @param {Object} settings - Settings object
 * @param {Object} settings.name - Name for grid
 * @param {Object} settings.cellSize - Cell size for grid in pixels
 * @returns SpatialHashGrid
 */
bento.define('modules/spatialhashgrid', [
    'bento/math/vector2',
    'bento/utils'
], function (Vector2, Utils) {
    'use strict';
    var floor = Math.floor;
    return function (settings) {
        var grid = {},
            name = settings.name || '',
            cellSize = settings.cellSize || new Vector2(16, 16),
            invCellSize = new Vector2(1 / cellSize.x, 1 / cellSize.y),
            posToKey = function (x, y) {
                return floor(x * invCellSize.x).toString() + ':' + floor(y * invCellSize.y);
            },
            boxToKeys = function (box) {
                var minX,
                    maxX,
                    minY,
                    maxY,
                    x,
                    y,
                    key,
                    keys = [];

                minX = floor(box.x * invCellSize.x);
                maxX = floor((box.getX2() - 1) * invCellSize.x);
                minY = floor(box.y * invCellSize.y);
                maxY = floor((box.getY2() - 1) * invCellSize.y);

                // loop through cells
                for (x = minX; x <= maxX; ++x) {
                    for (y = minY; y <= maxY; ++y) {
                        key = x.toString() + ':' + y.toString();
                        keys.push(key);
                    }
                }
                return keys;
            };

        // public
        return {
            timer: 0,
            /**
             * Inserts entity to grid
             * @function
             * @instance
             * @param {Entity} entity - Entity to add to grid
             * @name insert
             */
            insert: function (entity) {
                var box = entity.getBoundingBox(),
                    keys = boxToKeys(box),
                    key,
                    i;

                for (i = 0; i < keys.length; ++i) {
                    key = keys[i];
                    // initialize cell if needed
                    grid[key] = grid[key] || {};
                    // add
                    grid[key][entity.id] = entity;
                }
                // save grid positions to entity
                entity.shg = entity.shg || {
                    keys: {}
                };
                entity.shg.keys[name] = keys;
            },
            /**
             * Removes entity from grid
             * @function
             * @instance
             * @param {Entity} entity - Entity to remove
             * @name remove
             */
            remove: function (entity) {
                var i,
                    l,
                    shg = entity.shg,
                    keys,
                    key;

                // remove based on saved keys
                if (!shg || !shg.keys[name]) {
                    return;
                }
                keys = shg.keys[name];
                for (i = 0, l = keys.length; i < l; ++i) {
                    key = keys[i];
                    delete grid[key][entity.id];
                }
                entity.shg.keys[name] = null;
            },
            /**
             * Find hash collisions on grid (does not perform precise collision)
             * @function
             * @instance
             * @param {Entity} entity - Entity to check collision with
             * @param {Vector2} offset - Hypothetical offset of entity
             * @param {Function} [callback] - Called if possible collision occurs
             * @returns {Boolean} True of any collision
             * @name collidesWith
             */
            collidesWith: function (entity, offset, callback, deep) {
                var box = entity.getBoundingBox(),
                    keys = boxToKeys(box),
                    key,
                    i,
                    j,
                    cell,
                    hasCollided = false,
                    other;

                for (i = 0; i < keys.length; ++i) {
                    key = keys[i];
                    cell = grid[key];
                    if (cell) {
                        for (other in cell) {
                            if (!cell.hasOwnProperty(other)) {
                                continue;
                            }
                            // deep collision check?
                            if (deep) {
                                entity.collidesWith(cell[other], offset, callback);
                            } else if (callback) {
                                callback(cell[other]);
                            }
                        }
                        hasCollided = true;
                    }
                }

                return hasCollided;
            },
            /**
             * Clear grid
             * @function
             * @instance
             * @name reset
             */
            reset: function () {
                grid = {};
                // loop through all entities and reset their data??
            }
        };
    };
});