bento.define('singletons/firebase', [
    'bento',
    'bento/utils'
], function (
    Bento,
    Utils
) {
    'use strict';
    var initialized = false;
    var firebaseId = '';
    var acquireFirebaseId = function () {
        var fromSaveState = Bento.saveState.load('firebaseId', '');
        if (fromSaveState) {
            firebaseId = fromSaveState;
        } else if (window.FirebasePlugin && window.FirebasePlugin.getInstanceId) {
            window.FirebasePlugin.getInstanceId(function (token) {
                // Utils.checksum returns a 10 character string consisting of numbers. To make
                // it a bit more unique we'll take the first 10 characters of the original token
                firebaseId = 'id_' + token.substring(0, 10) + Utils.checksum(token);
                Bento.saveState.save('firebaseId', firebaseId);
            }, function (error) {
                console.error("ERROR: Couldn't acquire Firebase instance ID.");
                console.error(error);
            });
        } else {
            console.error("ERROR: Can't find FirebasePlugin.");
        }
    };
    /**
     * Checks if parameter values are either strings or numbers
     *
     * @param  {String/Number} value - current parameter value to be checked
     * @return {Boolean} returns true if value is either string or number, false otherwise
     */
    var isNumberOrString = function (value) {
        if (typeof value !== 'string' && typeof value !== 'number') {
            return false;
        }
        return true;
    };
    /**
     * Checks event and parameter name lengths against Firebase's requirements.
     * Event names can be up to 32 characters, parameter names up to 24, and parameter values up to 36
     *
     * @param  {String} str - the name to check
     * @param  {Number} max - the allowed length of the string
     * @return {Boolean} returns false if string exceeds maximum length, true otherwise
     */
    var checkStringLength = function (str, max) {
        if (str.length > max) {
            return false;
        }
        return true;
    };
    /**
     * Checks characters in event and parameter names against Firebase's requirements.
     * All names should start with a letter and contain only alphanumeric characters or underscores.
     *
     * @param  {String} str - the name to check
     * @return {Boolean} returns true if everything is correct, false otherwise
     */
    var checkForIllegalCharacters = function (str) {
        var code = 0;
        var i = 0;
        var il = 0;

        // event and parameter names should start with an alphabetic character
        code = str.charCodeAt(0);
        if (!(code > 64 && code < 91) && // upper alpha (A-Z)
            !(code > 96 && code < 123) // lower alpha (a-z)
        ) {
            return false;
        }

        // event and parameter names should only contain alphanumeric characters and underscores
        for (i = 0, il = str.length; i < il; ++i) {
            code = str.charCodeAt(i);
            if (!(code > 47 && code < 58) && // numeric (0-9)
                !(code > 64 && code < 91) && // upper alpha (A-Z)
                !(code > 96 && code < 123) && // lower alpha (a-z)
                code !== 95 // underscore (_)
            ) {
                return false;
            }
        }

        return true;
    };
    /**
     * Examines the event name against a list of reserved event names.
     *
     * Reserved names are a combination of these two lists:
     * https://support.google.com/firebase/answer/6317485
     * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Event
     *
     * @param  {String} str - the event name
     * @return {Boolean} returns false if the event name that is used is reserved, true otherwise
     */
    var checkForReservedNames = function (str) {
        var reserved = [
            'app_clear_data',
            'app_exception',
            'app_remove',
            'app_uninstall',
            'app_update',
            'dynamic_link_first_open',
            'dynamic_link_app_open',
            'dynamic_link_app_update',
            'error',
            'first_open',
            'in_app_purchase',
            'notification_dismiss',
            'notification_foreground',
            'notification_open',
            'notification_receive',
            'os_update',
            'session_start',
            'user_engagement'
        ];
        var i = 0;
        var il = 0;

        // the 'firebase_' prefix is reserved
        if (str.indexOf('firebase_') === 0) {
            return false;
        }

        for (i = 0, il = reserved.length; i < il; ++i) {
            if (str === reserved[i]) {
                return false;
            }
        }

        return true;
    };
    /**
     * Checks event name, parameter names, and parameter values against Firebase's requirements.
     * This function is currently only called when in dev mode.
     *
     * More info:
     * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Event
     * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Param
     *
     * @param  {String} name - Name of the event
     * @param  {Object} parameters - Event parameters
     * @return {Boolean} returns true if everything is correct, false otherwise
     */
    var inspectEvent = function (name, parameters) {
        // how long the strings are allowed to be for all individual components
        var maxLengths = {
            eventName: 32,
            paramName: 24,
            paramValue: 36
        };
        var amountOfParameters = 0;

        var key = '';

        // check event name against Firebase's requirements
        if (!checkStringLength(name, maxLengths.eventName)) {
            console.error("ERROR: Event name is too long: " + name);
            return false;
        }
        if (!checkForIllegalCharacters(name)) {
            console.error("ERROR: Illegal event name: " + name);
            return false;
        }
        if (!checkForReservedNames(name)) {
            console.error("ERROR: Using reserved event name: " + name);
            return false;
        }

        // check parameters against Firebase's requirements
        for (key in parameters) {
            if (!parameters.hasOwnProperty(key)) {
                continue;
            }
            // keep count of the amount of parameters
            ++amountOfParameters;

            if (!checkStringLength(key, maxLengths.paramName)) {
                console.error("ERROR: Parameter name is too long: " + name + " - " + key);
                return false;
            }
            if (!checkForIllegalCharacters(key)) {
                console.error("ERROR: Illegal parameter name: " + name + " - " + key);
                return false;
            }
            if (!isNumberOrString(parameters[key])) {
                console.error("ERROR: Parameter is not a string or number: " + name + " - " + key);
                return false;
            }
            // only check length of parameter value if it's a string because
            // checkStringLength uses String.length, which doesn't work on numbers
            if (typeof parameters[key] === 'string') {
                if (!checkStringLength(parameters[key], maxLengths.paramValue)) {
                    console.error("ERROR: Parameter value is too long: " + name + " - " + key);
                    return false;
                }
            }
        }

        // the maximum amount of unique parameters is 25
        if (amountOfParameters > 25) {
            console.error("ERROR: Too much parameters in " + name + ": " + amountOfParameters);
            return false;
        }

        return true;
    };
    var firebase = {
        /**
         * Only acquires Firebase instance ID for now
         */
        init: function () {
            if (window.FirebasePlugin) {
                initialized = true;
                acquireFirebaseId();
            } else {
                console.error("ERROR: Can't find FirebasePlugin.");
            }
        },
        /**
         * Send an event to Firebase analytics
         * @param  {String} name - Name of the event
         * @param  {Object} parameters - Event parameters. Optional
         */
        logEvent: function (name, parameters) {
            if (!initialized) {
                return;
            }
            if (!name) {
                console.error("ERROR: Event must have a name.");
                return;
            }
            // acquire firebase instance ID in case player had no internet connection until now
            if (!firebaseId) {
                acquireFirebaseId();
            }

            parameters = parameters || {};
            var internallyReserved = [
                'dev',
                'id',
                'version'
            ];
            var i = 0;
            var il = 0;

            for (i = 0, il = internallyReserved.length; i < il; ++i) {
                if (parameters.hasOwnProperty(internallyReserved[i])) {
                    console.error("ERROR: Don't use internally reserved parameter: " + internallyReserved[i]);
                    return;
                }
            }

            // identify the current player
            parameters['id'] = firebaseId;
            parameters['version'] = (Utils.isNativeIos() ? 'ios_' : 'and_') + Utils.version;
            parameters['dev'] = 0;

            if (Bento.isDev()) {
                // if event is sent during development mark it as such so it can be ignored later on
                parameters['dev'] = 1;

                // check if event names, parameter names, and parameter values are all
                // according to Firebase's specifications
                if (!inspectEvent(name, parameters)) {
                    return;
                }
            }

            window.FirebasePlugin.logEvent(
                name, // event name
                parameters, // event parameters
                function () { // success callback

                },
                function (e) { // error callback
                    console.error("ERROR: FirebasePlugin.logEvent failed: " + e);
                }
            );
        },
        /**
         * Standard events. Use standard parameters for these events. To find out which ones:
         * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Event
         *
         * @type {Object}
         */
        Events: {
            ADD_PAYMENT_INFO: 'add_payment_info', // This event signifies that a user has submitted their payment information to your app.
            ADD_TO_CART: 'add_to_cart', // This event signifies that an item was added to a cart for purchase. Add this event to a funnel with ECOMMERCE_PURCHASE to gauge the effectiveness of your checkout process. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            ADD_TO_WISHLIST: 'add_to_wishlist', // This event signifies that an item was added to a wishlist. Use this event to identify popular gift items in your app. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            APP_OPEN: 'app_open', // By logging this event when an App is moved to the foreground, developers can understand how often users leave and return during the course of a Session. Although Sessions are automatically reported, this event can provide further clarification around the continuous engagement of app-users.
            BEGIN_CHECKOUT: 'begin_checkout', // This event signifies that a user has begun the process of checking out. Add this event to a funnel with your ECOMMERCE_PURCHASE event to gauge the effectiveness of your checkout process. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            ECOMMERCE_PURCHASE: 'ecommerce_purchase', // This event signifies that an item was purchased by a user. Note: This is different from the in-app purchase event, which is reported automatically for Google Play-based apps. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            GENERATE_LEAD: 'generate_lead', // Log this event when a lead has been generated in the app to understand the efficacy of your install and re-engagement campaigns. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            JOIN_GROUP: 'join_group', // Log this event when a user joins a group such as a guild, team or family. Use this event to analyze how popular certain groups or social features are in your app.
            LEVEL_UP: 'level_up', // This event signifies that a player has leveled up in your gaming app. It can help you gauge the level distribution of your userbase and help you identify certain levels that are difficult to pass.
            LOGIN: 'login', // Apps with a login feature can report this event to signify that a user has logged in.
            POST_SCORE: 'post_score', // Log this event when the user posts a score in your gaming app. This event can help you understand how users are actually performing in your game and it can help you correlate high scores with certain audiences or behaviors.
            PRESENT_OFFER: 'present_offer', // This event signifies that the app has presented a purchase offer to a user. Add this event to a funnel with the ADD_TO_CART and ECOMMERCE_PURCHASE to gauge your conversion process. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            PURCHASE_REFUND: 'purchase_refund', // This event signifies that an item purchase was refunded. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            SEARCH: 'search', // Apps that support search features can use this event to contextualize search operations by supplying the appropriate, corresponding parameters. This event can help you identify the most popular content in your app.
            SELECT_CONTENT: 'select_content', // This general purpose event signifies that a user has selected some content of a certain type in an app. The content can be any object in your app. This event can help you identify popular content and categories of content in your app.
            SHARE: 'share', // Apps with social features can log the Share event to identify the most viral content.
            SIGN_UP: 'sign_up', // This event indicates that a user has signed up for an account in your app. The parameter signifies the method by which the user signed up. Use this event to understand the different behaviors between logged in and logged out users.
            SPEND_VIRTUAL_CURRENCY: 'spend_virtual_currency', // This event tracks the sale of virtual goods in your app and can help you identify which virtual goods are the most popular objects of purchase.
            TUTORIAL_BEGIN: 'tutorial_begin', // This event signifies the start of the on-boarding process in your app. Use this in a funnel with TUTORIAL_COMPLETE to understand how many users complete this process and move on to the full app experience.
            TUTORIAL_COMPLETE: 'tutorial_complete', // Use this event to signify the user’s completion of your app’s on-boarding process. Add this to a funnel with TUTORIAL_BEGIN to gauge the completion rate of your on-boarding process.
            UNLOCK_ACHIEVEMENT: 'unlock_achievement', // Log this event when the user has unlocked an achievement in your game. Since achievements generally represent the breadth of a gaming experience, this event can help you understand how many users are experiencing all that your game has to offer.
            VIEW_ITEM: 'view_item', // This event signifies that some content was shown to the user. This content may be a product, a webpage or just a simple image or text. Use the appropriate parameters to contextualize the event. Use this event to discover the most popular items viewed in your app. Note: If you supply the VALUE parameter, you must also supply the CURRENCY parameter so that revenue metrics can be computed accurately.
            VIEW_ITEM_LIST: 'view_item_list', // Log this event when the user has been presented with a list of items of a certain category.
            VIEW_SEARCH_RESULTS: 'view_search_results' // Log this event when the user has been presented with the results of a search.
        },
        /**
         * Standard parameters. For use with standard events, see for more info:
         * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Param
         *
         * @type {Object}
         */
        Parameters: {
            ACHIEVEMENT_ID: 'achievement_id', // Game achievement ID (String).
            CHARACTER: 'character', // Character used in game (String).
            CONTENT_TYPE: 'content_type', // Type of content selected (String).
            COUPON: 'coupon', // Coupon code for a purchasable item (String).
            CURRENCY: 'currency', // Purchase currency in 3 letter ISO_4217 format (String).
            DESTINATION: 'destination', // Flight or Travel destination (String).
            END_DATE: 'end_date', // The arrival date, check-out date, or rental end date for the item (String).
            FLIGHT_NUMBER: 'flight_number', // Flight number for travel events (String).
            GROUP_ID: 'group_id', // Group/clan/guild id (String).
            ITEM_CATEGORY: 'item_category', // Item category (String).
            ITEM_ID: 'item_id', // Item ID (String).
            ITEM_LOCATION_ID: 'item_location_id', // The Google Place ID that corresponds to the associated item (String).
            ITEM_NAME: 'item_name', // Item name (String).
            LEVEL: 'level', // Level in game (long).
            LOCATION: 'location', // Location (String).
            NUMBER_OF_NIGHTS: 'number_of_nights', // Number of nights staying at hotel (long).
            NUMBER_OF_PASSENGERS: 'number_of_passengers', // Number of passengers traveling (long).
            NUMBER_OF_ROOMS: 'number_of_rooms', // Number of rooms for travel events (long).
            ORIGIN: 'origin', // Flight or Travel origin (String).
            PRICE: 'price', // Purchase price (double).
            PRODUCT_CATEGORY: 'product_category', // Product category (String).
            PRODUCT_ID: 'product_id', // Purchase product ID (String).
            PRODUCT_NAME: 'product_name', // Purchase item name (String).
            QUANTITY: 'quantity', // Purchase quantity (long).
            SCORE: 'score', // Score in game (long).
            SEARCH_TERM: 'search_term', // The search string/keywords used (String).
            SHIPPING: 'shipping', // Shipping cost (double).
            SIGN_UP_METHOD: 'sign_up_method', // Signup method (String).
            START_DATE: 'start_date', // The departure date, check-in date, or rental start date for the item (String).
            TAX: 'tax', // Tax amount (double).
            TRANSACTION_ID: 'transaction_id', // A single ID for a ecommerce group transaction (String).
            TRAVEL_CLASS: 'travel_class', // Travel class (String).
            VALUE: 'value', // A context-specific numeric value which is accumulated automatically for each event type.
            VIRTUAL_CURRENCY_NAME: 'virtual_currency_name' // Name of virtual currency type (String).
        }
    };

    return firebase;
});