/**
 * Cocoon's notification plugin
 * http://ludei.github.io/atomic-plugins-docs/dist/doc/js/Cocoon.Notification.html
 */
bento.define('singletons/notification', [
    'bento',
    'bento/utils',
    'gui/dialog',
    'gui/dailygift'
], function (
    Bento,
    Utils,
    Dialog,
    DailyGift
) {
    'use strict';
    var initialized = false;
    var localService;
    var nothing = function () {};
    var allowed;
    /**
     * Initialize local notification system. Registers service automatically after initialization.
     * Registering may show a Dialog to request user permissions.
     * @param {Boolean} onComplete - Optional callback to be called after successful initialization
     */
    var init = function (onComplete) {
        // check if the plugin is available
        if (!Cocoon || !Cocoon.Notification || !Cocoon.Notification.Local) {
            if (Bento.isDev()) {
                throw "ERROR: No local notification plugin found";
            } else {
                console.log("ERROR: No local notification plugin found");
            }
            return;
        }
        if (!initialized) {
            localService = Cocoon.Notification.Local;

            /**
             * Notification listener
             * @param  {Object} userData - Optional object through which properties and methods can be passed
             *             {Function} onComplete: optional callback to be called when app is started
             */
            localService.on('notification', function (userData) {
                if (userData.dailyGift) {
                    // if not in galaxy or overworld screen, don't show daily gift just yet
                    if (Bento.screens.getCurrentScreen().name.indexOf('map') < 0)
                        return;
                    // small delay to wait for the fade
                    Utils.timeout(
                        30,
                        function () {
                            // should give daily gift?
                            if (DailyGift.canGive()) {
                                new DailyGift({});
                            }
                        },
                        Utils.PauseLevels.ALWAYS
                    );
                }
            });

            localService.initialize({}, function (registered, error) {
                localService.isRegistered(function (granted) {
                    if (!granted) {
                        if (Bento.isDev()) {
                            throw "ERROR: couldn't register local notifications";
                        } else {
                            console.log("WARNING: couldn't register local notifications");
                        }
                    } else {
                        console.log('Local notification registered successfully');
                    }

                    initialized = granted;

                    if (onComplete)
                        onComplete();
                });
            });

            return;
        } else {
            if (Bento.isDev()) {
                throw 'ERROR: Local notifications already initialized.';
            } else {
                console.log('WARNING: Local notifications already initialized.');
            }
        }
    };

    /**
     * Sends/schedules notification
     * @param  {Object} notification - Notification to send/schedule
     *             {String} title: (tested on iOS) The notification title. By default, it will be the app title
     *             {String} message: The notification message. By default, it will be empty
     *             {Boolean} soundEnabled: A flag that indicates if the sound should be enabled for the notification. By default, it will be true
     *             {Number} badgeNumber: (iOS only it seems) The number that will appear in the badge of the application icon in the home screen. By default, it will be 0
     *             {Object} userData: The JSON data to attached to the notification. By default, it will be empty
     *             {String} contentBody: (haven't yet seen this appear) The body content to be showed in the expanded notification information. By default, it will be empty
     *             {String} contentTitle: (haven't yet seen this appear) The title to be showed in the expanded notification information. By default, it will be empty
     *             {Number} date: Time in millisecs from 1970 when the notification will be fired. By default, it will be 1 second (1000)
     *             {String} icon: (Optional) (Android only) Name of the custom icon that you have bundled in the application as a resource (You will need a custom Cordova plugin to bundle resources in the app). In Android is the icon name without extension
     */
    var sendNotification = function (notification) {
        if (!notification) {
            if (Bento.isDev()) {
                throw 'ERROR: No notification to send';
            } else {
                console.log('WARNING: No notification to send');
            }
            return;
        }

        // cancel any outstanding notifications just in case
        localService.cancelAllNotifications();

        localService.send(notification, function () {
            console.log('Notification sent successfully');
        });
    };

    /**
     * Ask the player if they want to receive notifications
     * @param  {Object} notification - notification object to pass to sendNotification
     */
    var ask = function (notification, onComplete) {
        var timesAsked = Bento.saveState.load('timesAsked', 0);
        var oldPauseLevel = Bento.objects.isPaused();
        Bento.objects.pause(Utils.PauseLevels.EXIT);

        timesAsked++;

        // save that player has been asked to prevent asking player about notification dialog after each daily gift
        Bento.saveState.save('notificationsAsked', true);

        // show dialog asking if player they want notifications
        new Dialog({
            name: 'askNotification',
            text: Utils.Localization.getText('allowNotif'),
            pauseLevel: Utils.PauseLevels.EXIT,
            type: 'yesno',
            // initialize notifications which prompts a permission dialog from the OS
            onYes: function () {
                allowed = true;
                Bento.saveState.save('notificationsAllowed', allowed);

                Bento.saveState.save('timesAsked', timesAsked);
                Utils.Analytics.event('Notifications', 'allowAfter_' + timesAsked);
                Utils.Firebase.logEvent('notfication_allowed', {
                    value: 1,
                    times_asked: timesAsked
                });

                Bento.objects.pause(oldPauseLevel);
                module.sendNotification(notification);
            },
            // just go back to the previous state of the game
            onNo: function () {
                Bento.saveState.save('timesAsked', timesAsked);
                Utils.Analytics.event('Notifications', 'denyFor_' + timesAsked);
                Utils.Firebase.logEvent('notfication_denied', {
                    value: 1,
                    times_asked: timesAsked
                });

                Bento.objects.pause(oldPauseLevel);
            },
            onClose: onComplete
        });

        return;
    };

    var module = {
        init: init,
        // TODO: onComplete is now only used when player is asked about receiving notifications
        // This is because it was added as a quick fix to stop the Twitter tutorial from popping up
        // in between the daily gift and the notification request
        sendNotification: function (notification, onComplete) {
            if (Utils.WEB) {
                return;
            }
            allowed = Bento.saveState.load('notificationsAllowed', false);
            if (!allowed) {
                ask(notification, onComplete);

                return;
            }
            if (!initialized) {
                // if not initialized yet, initialize and send/schedule notification immediately afterwards
                init(function () {
                    sendNotification(notification);
                });

                return;
            } else {
                sendNotification(notification);
            }
        },
        resetBadgeNumber: function () {
            if (!initialized) {
                console.log('Local Notifications are not initialized');
                return;
            }
            // reset notification number on app icon
            if (Utils.isNativeIos()) {
                localService.setBadgeNumber(0);
            }
        }
    };

    if (!window.Cocoon) {
        module.init = nothing;
        module.sendNotification = nothing;
        module.resetBadgeNumber = nothing;
    }

    return module;
});