/**
 * Reads Tiled JSON file and spawns entities accordingly.
 * Backgrounds are merged into a canvas image
 * Optimized for loading smaller parts of a single map
 * <br>Exports: Function
 * @module bento/tiled
 * @param {Object} settings - Settings object
 * @param {String} settings.name - Asset name of the json file
 * @param {Boolean} [settings.spawn] - Spawns entities
 * @returns Object
 */
bento.define('subtiled', [
    'bento',
    'bento/entity',
    'bento/eventsystem',
    'bento/components/sprite',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/math/polygon',
    'bento/packedimage',
    'bento/utils',
    'bento/objectpool',
    'entities/tilecolliders'
], function (
    Bento,
    Entity,
    EventSystem,
    Sprite,
    Vector2,
    Rectangle,
    Polygon,
    PackedImage,
    Utils,
    ObjectPool,
    TileColliders
) {
    'use strict';
    var canvasSize = new Vector2(256, 256),
        canvasPool = new ObjectPool({
            poolSize: 20,
            constructor: function () {
                var canvas = document.createElement('canvas');
                canvas.width = canvasSize.width;
                canvas.height = canvasSize.height;

                return canvas;
            },
            destructor: function (obj) {
                // clear canvas
                var context = obj.getContext('2d');
                context.clearRect(0, 0, obj.width, obj.height);
                // clear texture
                if (obj.texture) {
                    obj.texture = null;
                }
                return obj;
            }
        });
    // cached layer canvasses
    // id by rectangle x, y, layer
    var cachedLayers = {};
    EventSystem.on('clearTiledCache', function () {
        cachedLayers = {};
    });

    return function (settings, onReady) {
        /*settings = {
            json: Json, // Asset
            name: String, // name of JSON file
            background: Boolean // TODO false: splits tileLayer tile entities,
            spawn: Boolean // adds objects into game immediately
            rectangle: Rectangle //area to load (dimension is tiles, not pixels),
            offset: Vector2 // offset all positions
            onLoad: Function // called when tiled is loaded,
            cachedLayers: Boolean // use cache?
        }*/
        var json = settings.json,
            offset = settings.offset || new Vector2(0, 0),
            width = json.width,
            height = json.height,
            tileWidth = json.tilewidth,
            tileHeight = json.tileheight,
            tilesets = json.tilesets,
            tileLayerImages = [],
            tileDraws = 0,
            spawnTileObjects = json.properties ? json.properties.spawnTiles : false,
            /*{
                canvas: Canvas,
                context: Context,
                packedImage: PackedImage,
                entity: Entity
            }*/
            // canvas = document.createElement('canvas'),
            // context = canvas.getContext('2d'),
            objects = [],
            shapes = [],
            viewport = Bento.getViewport(),
            pixelRectangle = new Rectangle(0, 0, 0, 0),
            packedImage,
            background,
            tileColliders,
            // loaded
            loadedEntities = 0,
            totalEntities = 0,
            /**
             * Check if all entities are loaded
             */
            checkLoaded = function () {
                // console.log(loadedEntities, totalEntities)
                if (loadedEntities === totalEntities && settings.onLoad) {
                    settings.onLoad(settings, objects, tileLayerImages);
                }
            },
            /**
             * Loop through tilesets and tileset to be imported
             */
            importTilesets = function () {
                var i, l,
                    tilesets = json.tilesets,
                    tileset,
                    source,
                    sourceFolders,
                    externalJson,
                    folders;

                // loop through all tilesets
                for (i = 0, l = tilesets.length; i < l; ++i) {
                    tileset = tilesets[i];
                    source = tileset.source;
                    if (source && source.indexOf('../') >= 0) {
                        // ignore file extension
                        source = source.replace('.tsx', '');
                        source = source.replace('.json', '');
                        // extract filename
                        if (source.indexOf('../') >= 0) {
                            // remove ../
                            while (source.indexOf('../') >= 0)
                                source = source.replace('../', '');
                            // relative folder, start with tiled json name
                            externalJson = settings.name;
                            // split folders
                            folders = externalJson.split('/');
                            // remove json filename
                            folders.pop();
                            // replace last folder with source
                            folders[folders.length - 1] = source;
                            // recombine
                            externalJson = folders.join('/');
                        } else {
                            externalJson = source;
                        }

                        // swap out tileset
                        // tilesets[i] = Bento.assets.getJson(externalJson);

                        // clone the object, setting the firstgid screws up other maps
                        // that share the same tileset
                        // TODO: make this more efficient
                        tilesets[i] = JSON.parse(JSON.stringify(Bento.assets.getJson(externalJson)));
                        tilesets[i].firstgid = tileset.firstgid;
                        tilesets[i].source = tileset.source.replace('../', '');
                    }
                }
            },
            getTileset = function (gid) {
                var l,
                    tileset,
                    current = null;
                // loop through tilesets and find the highest firstgid that's
                // still lower or equal to the gid
                for (l = 0; l < tilesets.length; ++l) {
                    tileset = tilesets[l];
                    if (tileset.firstgid <= gid) {
                        current = tileset;
                    }
                }
                return current;
            },
            getTile = function (tileset, gid) {
                var index,
                    tilesetWidth,
                    tilesetHeight;
                if (tileset === null) {
                    return null;
                }
                index = gid - tileset.firstgid;
                tilesetWidth = Math.floor(tileset.imagewidth / tileset.tilewidth);
                tilesetHeight = Math.floor(tileset.imageheight / tileset.tileheight);
                return {
                    // convention: the tileset name must be equal to the asset name!
                    subimage: Bento.assets.getImage(tileset.name),
                    x: (index % tilesetWidth) * tileset.tilewidth,
                    y: Math.floor(index / tilesetWidth) * tileset.tileheight,
                    width: tileset.tilewidth,
                    height: tileset.tileheight
                };
            },
            drawTile = function (layer, x, y, target) {
                var gid = layer.data[y * width + x],
                    // get correct tileset and image
                    tileset = gid === 0 ? null : getTileset(gid),
                    tile = gid === 0 ? null : getTile(tileset, gid),
                    subImage,
                    tileproperties,
                    tileproperty;

                if (gid === 0) {
                    return;
                }

                if (tile) {
                    subImage = tile.subimage;

                    // spawn rectangle?
                    if (true || spawnTileObjects) {
                        // search properties
                        tileproperties = tileset.tileproperties;
                        if (tileproperties) {
                            tileproperty = tileproperties[gid - tileset.firstgid];
                            if (tileproperty) {
                                // found properties!
                                if (tileproperty.type === 'solid') {
                                    // spawns a colliding tile here
                                    spawnShape(new Rectangle(
                                        offset.x + x * tileWidth,
                                        offset.y + y * tileHeight,
                                        tileWidth,
                                        tileHeight
                                    ), {
                                        type: 'solid'
                                    }, true);
                                }
                                if (tileproperty.type === 'spikes') {
                                    // spawns a colliding tile here
                                    spawnShape(new Rectangle(
                                        offset.x + x * tileWidth,
                                        offset.y + y * tileHeight,
                                        tileWidth,
                                        tileHeight
                                    ), {
                                        type: 'spikes'
                                    }, true);
                                }
                                if (tileproperty.module) {
                                    // spawns a module here
                                    spawn(
                                        tileproperty.module, {
                                            x: offset.x + x * tileWidth,
                                            y: offset.y + y * tileHeight
                                        },
                                        tileproperty
                                    );
                                }

                                if (tileproperty.noTile) {
                                    // do not draw!
                                    return;
                                }
                            }
                        }
                    }
                    if (layer.properties && layer.properties.noTile) {
                        // do not draw!
                        return;
                    }

                    // offset to canvas corner
                    x -= settings.rectangle.x;
                    y -= settings.rectangle.y;
                    // METHOD 1: predraw background to offscreen canvas
                    target.globalAlpha = layer.opacity || 1;
                    target.drawImage(
                        subImage.image,
                        subImage.x + tile.x,
                        subImage.y + tile.y,
                        tile.width,
                        tile.height,
                        x * tileWidth,
                        y * tileHeight,
                        tileWidth,
                        tileHeight
                    );
                    tileDraws += 1;
                }

            },
            spawn = function (name, obj, tilesetProperties) {
                var x = obj.x,
                    y = obj.y,
                    params = {
                        tiled: {
                            position: new Vector2(offset.x + x, offset.y + y),
                            tile: tilesetProperties,
                            object: Utils.extend(obj, obj.properties), // extend properties with tiled info
                            json: settings.name // reference to current json name
                        }
                    };
                totalEntities += 1;

                // collect parameters
                // Utils.extend(params, tilesetProperties);
                // Utils.extend(params, obj.properties);

                bento.require([name], function (Instance) {
                    // console.log('spawn', name)
                    // var instance = Instance.apply(this, [params]),
                    var instance = new Instance(params),
                        origin = instance.origin,
                        dimension = instance.dimension,
                        prop,
                        addProperties = function (properties) {
                            var prop;
                            for (prop in properties) {
                                if (prop === 'module' || prop.match(/param\d+/)) {
                                    continue;
                                }
                                if (properties.hasOwnProperty(prop)) {
                                    // number or string?
                                    if (isNaN(properties[prop])) {
                                        instance[prop] = properties[prop];
                                    } else {
                                        instance[prop] = (+properties[prop]);
                                    }
                                }
                            }
                        };

                    instance.position = new Vector2(offset.x + x + origin.x, offset.y + y + (origin.y - dimension.height));

                    // add in tileset properties
                    //addProperties(tilesetProperties);
                    // add tile properties
                    //addProperties(obj.properties);

                    // add to game
                    if (settings.spawn) {
                        Bento.objects.add(instance);
                    }
                    objects.push(instance);
                    loadedEntities += 1;
                    checkLoaded();
                });
            },
            spawnObject = function (obj) {
                var gid = obj.gid,
                    // get tileset: should contain module name
                    tileset = getTileset(gid),
                    id = gid - tileset.firstgid,
                    properties,
                    moduleName;
                if (tileset.tileproperties) {
                    properties = tileset.tileproperties[id.toString()];
                    if (properties) {
                        moduleName = properties.module;
                    }
                }
                if (moduleName) {
                    spawn(moduleName, obj, properties);
                }
            },
            spawnShape = function (shape, tileObj, ground) {
                var obj;
                if (settings.spawn) {
                    obj = new Entity({
                        z: 0,
                        name: tileObj.type,
                        family: [tileObj.type],
                        useHshg: true,
                        staticHshg: true,
                        // pass tile properties and object properties
                        tiled: {
                            object: tileObj.properties || {},
                            tile: null
                        }
                    });
                    // remove update and draw functions to save processing power
                    obj.update = null;
                    obj.draw = null;

                    // debug
                    // obj.attach(new Debug({}));

                    obj.boundingBox = shape;
                    if (!tileColliders) {
                        Bento.objects.get('tileColliders', function (obj) {
                            tileColliders = obj;
                        });
                    }
                    if (!tileColliders) {
                        tileColliders = new TileColliders({});
                        Bento.objects.attach(tileColliders);
                    }
                    if (ground && tileColliders) {
                        tileColliders.insert(obj);
                        obj.inHash = true;
                    } else {
                        Bento.objects.add(obj);
                    }
                }
                shape.type = tileObj.type;
                shapes.push(obj);
            },
            addLayer = function (index, cachedCanvas) {
                var canvas, context;
                var packedImage,
                    entity,
                    layer;

                if (Utils.isDefined(cachedCanvas)) {
                    canvas = cachedCanvas;
                } else {
                    // canvas = canvasPool.get();
                    canvas = document.createElement('canvas');
                    canvas.width = canvasSize.width;
                    canvas.height = canvasSize.height;
                }
                context = canvas.getContext('2d');

                // setup packedimage and entity
                packedImage = new PackedImage(canvas);
                entity = new Entity({
                    z: 0,
                    name: 'tileLayer',
                    position: new Vector2(offset.x + settings.rectangle.x * tileWidth, offset.y + settings.rectangle.y * tileHeight),
                    origin: new Vector2(0, 0),
                    components: [
                        new Sprite({
                            image: packedImage
                        })
                    ],
                    family: ['tileLayers']
                }).attach({
                    destroy: function (data) {
                        // put canvas back
                        // canvasPool.discard(canvas);
                        // clear references
                        packedImage = undefined;
                        layer = undefined;
                        canvas = undefined;
                        context = undefined;
                        entity = undefined;
                    }
                });
                // collect
                layer = {
                    canvas: canvas,
                    context: context,
                    packedImage: new PackedImage(canvas),
                    entity: entity
                };

                // add background to game
                if (settings.spawn) {
                    Bento.objects.add(entity);
                }
                tileLayerImages[index] = layer;
                return layer;

            },
            loopLayers = function () {
                var i, il,
                    j, jl,
                    k, kl,
                    layers = json.layers,
                    layer,
                    lh = layers.height,
                    lw = layers.width,
                    object,
                    points,
                    properties,
                    layerId = 0,
                    type;

                // check cache
                var id, cache;

                // loop through layers
                for (k = 0, kl = layers.length; k < kl; ++k) {
                    layer = layers[k];
                    type = layer.type;
                    if (type === 'tilelayer') {
                        if (!layer.visible) {
                            continue;
                        }
                        // was layer drawn before?
                        if (settings.cachedLayers) {
                            cache = null;
                            if (settings.cachedLayers) {
                                id = settings.rectangle.x + ',' + settings.rectangle.y + ',' + k;
                                cache = cachedLayers[id];
                            }
                        }

                        // spawn layer if needed
                        properties = layer.properties;
                        if (properties) {
                            layerId = properties.layer || 0;
                        } else {
                            layerId = 0;
                        }
                        if (!tileLayerImages[layerId]) {
                            if (settings.cachedLayers) {
                                if (cache) {
                                    // layer was drawn before!
                                    addLayer(layerId, cache);
                                    continue;
                                } else {
                                    // get ready to store it
                                    cachedLayers[id] = addLayer(layerId).canvas;
                                }
                            } else {
                                addLayer(layerId);
                            }
                        }

                        // loop through tiles
                        jl = Math.min(settings.rectangle.getY2(), height);

                        // add extra row in background layer
                        if (layerId === 0) {
                            jl = Math.min(settings.rectangle.getY2() + 1, height);
                        }
                        for (j = settings.rectangle.y; j < jl; ++j) {
                            il = Math.min(settings.rectangle.getX2(), width);

                            // add extra row in background layer
                            if (layerId === 0) {
                                il = Math.min(settings.rectangle.getX2() + 1, width);
                            }
                            for (i = settings.rectangle.x; i < il; ++i) {
                                // out of bounds
                                if (j < 0 || i < 0 || j >= lh || i >= lw) {
                                    continue;
                                }
                                drawTile(layer, i, j, tileLayerImages[layerId].context);
                            }
                        }
                    } else if (type === 'objectgroup') {
                        il = layer.objects.length;
                        for (i = 0; i < il; ++i) {
                            object = layer.objects[i];

                            // object is not in rectangle
                            if (!pixelRectangle.hasPosition(new Vector2(object.x, object.y))) {
                                continue;
                            }

                            // default type is solid
                            if (object.type === '') {
                                object.type = 'solid';
                            }

                            if (object.gid) {
                                // normal object
                                spawnObject(object);
                            } else if (object.polygon) {
                                // polygon
                                points = [];
                                for (j = 0; j < object.polygon.length; ++j) {
                                    points.push({
                                        x: offset.x + object.polygon[j].x + object.x,
                                        y: offset.y + object.polygon[j].y + object.y + 1
                                    });
                                    // shift polygons 1 pixel down?
                                    // something might be wrong with polygon definition
                                }
                                spawnShape(new Polygon(points), object);
                            } else {
                                if (object.properties.module) {
                                    // special rectangle
                                    spawn(object.properties.module, object, {});
                                } else {
                                    // rectangle
                                    spawnShape(new Rectangle(offset.x + object.x, offset.y + object.y, object.width, object.height), object);
                                }
                            }
                        }
                    }
                }
                tileDraws = 0;
            };

        // import external tileset
        importTilesets();

        // calculate pixel rectangle
        pixelRectangle.x = settings.rectangle.x * tileWidth;
        pixelRectangle.y = settings.rectangle.y * tileHeight;
        pixelRectangle.width = settings.rectangle.width * tileWidth;
        pixelRectangle.height = settings.rectangle.height * tileHeight;
        // update width and height of canvasPool
        canvasSize = pixelRectangle.clone();

        loopLayers();
        checkLoaded();

        // at least 1 canvas for the background
        // background = addLayer().entity;
        if (tileLayerImages.length) {
            background = tileLayerImages[0].entity;
        }

        canvasPool.init();

        return {
            clearCache: function () {
                cachedLayers = {};
            },
            /**
             * All tilelayers merged into one entity
             * @instance
             * @name tileLayer
             */
            tileLayer: background,
            /**
             * Other tilelayers (includes background)
             * @instance
             * @name tileLayers
             */
            tileLayers: tileLayerImages,
            /**
             * Array of entities
             * @instance
             * @name objects
             */
            objects: objects,
            /**
             * Array of Rectangles and Polygons
             * @instance
             * @name shapes
             */
            shapes: shapes,
            /**
             * Size of the screen
             * @instance
             * @name dimension
             */
            dimension: new Rectangle(0, 0, tileWidth * width, tileHeight * height),
            /**
             * Offsets the entire object and its parts with the specified position.
             * @function
             * @instance
             * @name offset
             */
            offset: function (position) {
                for (var i = 0, len = shapes.length; i < len; i++) {
                    shapes[i].x += position.x;
                    shapes[i].y += position.y;
                }
                for (i = 0, len = objects.length; i < len; i++) {
                    objects[i].offset(position);
                }
                for (i = 0, len = tileLayerImages.length; i < len; i++) {
                    tileLayerImages[i].entity.position.addTo(position);
                }
            },
            /**
             * Removes the tileLayer, objects, and shapes
             * @function
             * @instance
             * @name remove
             */
            remove: function () {
                var shape;
                for (var i = 0, len = tileLayerImages.length; i < len; i++) {
                    Bento.objects.remove(tileLayerImages[i].entity);
                }
                tileLayerImages.length = 0;
                for (i = 0, len = shapes.length; i < len; i++) {
                    shape = shapes[i];
                    if (shapes.inHash) {
                        hashGround.remove(shape);
                    } else {
                        Bento.objects.remove(shape);
                    }
                }
                shapes.length = 0;
                for (i = 0, len = objects.length; i < len; i++) {
                    if (objects[i].doNotRemove) {
                        continue;
                    }
                    Bento.objects.remove(objects[i]);
                }
                objects.length = 0;
            }
        };
    };
});